import abc
import contextlib
import enum
import operator
import os
import sys
import time
from dataclasses import dataclass
from fnmatch import fnmatch
from math import ceil
from time import monotonic
from typing import Any, Dict, Iterator, List, Optional, Sequence, Tuple

import click
from click import style, unstyle
from yarl import URL

from neuromation.api import (
    AbstractDeleteProgress,
    AbstractRecursiveFileProgress,
    Action,
    FileStatus,
    FileStatusType,
    StorageProgressComplete,
    StorageProgressDelete,
    StorageProgressEnterDir,
    StorageProgressFail,
    StorageProgressLeaveDir,
    StorageProgressStart,
    StorageProgressStep,
)
from neuromation.api.url_utils import _extract_path
from neuromation.cli.printer import TTYPrinter
from neuromation.cli.root import Root
from neuromation.cli.utils import format_size


RECENT_TIME_DELTA = 365 * 24 * 60 * 60 / 2
TIME_FORMAT = "%Y-%m-%d %H:%M:%S"


def chunks(list: Sequence[Any], size: int) -> Sequence[Any]:
    result = []
    for i in range(0, len(list), size):
        result.append(list[i : i + size])
    return result


def transpose(columns: Sequence[Sequence[Any]]) -> Sequence[Sequence[Any]]:
    height = len(columns)
    width = len(columns[0])
    result: Sequence[List[Any]] = [[] for _ in range(width)]
    for i in range(width):
        for j in range(height):
            if i < len(columns[j]):
                result[i].append(columns[j][i])
    return result


class GnuIndicators(str, enum.Enum):
    LEFT = "lc"
    RIGHT = "rc"
    END = "ec"
    RESET = "rs"
    NORM = "no"
    FILE = "fi"
    DIR = "di"
    LINK = "ln"
    FIFO = "pi"
    SOCKET = "so"
    BLK = "bd"
    CHR = "cd"
    MISSING = "mi"
    ORPHAN = "or"
    EXEC = "ex"
    DOOR = "do"
    SETUID = "su"
    SETGID = "sg"
    STICKY = "st"
    OTHER_WRITABLE = "ow"
    STICKY_OTHER_WRITABLE = "tw"
    CAP = "ca"
    MULTI_HARD_LINK = "mh"
    CLR_TO_EOL = "cl"


class ParseState(enum.Enum):
    PS_START = enum.auto()
    PS_LEFT = enum.auto()
    PS_ESCAPED = enum.auto()
    PS_ESCAPED_END = enum.auto()
    PS_RIGHT = enum.auto()
    PS_OCTAL = enum.auto()
    PS_HEX = enum.auto()
    PS_CARRET = enum.auto()


class BasePainter(abc.ABC):
    @abc.abstractmethod
    def paint(self, label: str, type: FileStatusType) -> str:  # pragma: no cover
        pass


class NonePainter(BasePainter):
    def paint(self, label: str, type: FileStatusType) -> str:
        return label


class QuotedPainter(BasePainter):
    def paint(self, label: str, type: FileStatusType) -> str:
        if "'" not in label:
            return "'" + label + "'"
        else:
            return '"' + label + '"'


class GnuPainter(BasePainter):
    def __init__(self, ls_colors: str, *, underline: bool = False):
        self._defaults()
        self._parse_ls_colors(ls_colors)
        self._underline = underline

    def _defaults(self) -> None:
        self.color_indicator: Dict[GnuIndicators, str] = {
            GnuIndicators.LEFT: "\033[",
            GnuIndicators.RIGHT: "m",
            GnuIndicators.END: "",
            GnuIndicators.RESET: "0",
            GnuIndicators.NORM: "",
            GnuIndicators.FILE: "",
            GnuIndicators.DIR: "01;34",
            GnuIndicators.LINK: "01;36",
            GnuIndicators.FIFO: "33",
            GnuIndicators.SOCKET: "01;35",
            GnuIndicators.BLK: "01;33",
            GnuIndicators.CHR: "01;33",
            GnuIndicators.MISSING: "",
            GnuIndicators.ORPHAN: "",
            GnuIndicators.EXEC: "01;32",
            GnuIndicators.DOOR: "01;35",
            GnuIndicators.SETUID: "37;41",
            GnuIndicators.SETGID: "30;43",
            GnuIndicators.STICKY: "37;44",
            GnuIndicators.OTHER_WRITABLE: "34;42",
            GnuIndicators.STICKY_OTHER_WRITABLE: "30;42",
            GnuIndicators.CAP: "30;41",
            GnuIndicators.MULTI_HARD_LINK: "",
            GnuIndicators.CLR_TO_EOL: "\033[K",
        }
        self.color_ext_type: Dict[str, str] = {}

    def _parse_ls_colors(self, ls_colors: str) -> None:
        def process(left: str, right: str) -> None:
            try:
                self.color_indicator[GnuIndicators(left)] = right
            except ValueError:
                self.color_ext_type[left] = right

        pos = 0
        left = right = escaped = ""
        num = 0
        state = ParseState.PS_START
        stack: List[ParseState] = []
        while pos < len(ls_colors):
            char = ls_colors[pos]
            if state == ParseState.PS_START:
                if char == ":":  # ignore colon
                    pos += 1
                else:
                    left = ""
                    state = ParseState.PS_LEFT
            elif state == ParseState.PS_OCTAL:
                if char in ["0", "1", "2", "3", "4", "5", "6", "7"]:
                    num = num * 8 + ord(char) - ord("0")
                    if num > 7:
                        state = ParseState.PS_ESCAPED_END
                        escaped = chr(num)
                    pos += 1
                else:
                    state = ParseState.PS_ESCAPED_END
                    escaped = chr(num)
            elif state == ParseState.PS_HEX:
                if char in ["0", "1", "2", "3", "4", "5", "6", "7", "8", "9"]:
                    num = num * 16 + ord(char) - ord("0")
                    if num > 15:
                        state = ParseState.PS_ESCAPED_END
                        escaped = chr(num)
                    pos += 1
                elif char.upper() in ["A", "B", "C", "D", "E", "F"]:
                    num = num * 16 + 10 + ord(char.upper()) - ord("A")
                    if num > 15:
                        state = ParseState.PS_ESCAPED_END
                        escaped = chr(num)
                    pos += 1
                else:
                    state = ParseState.PS_ESCAPED_END
                    escaped = chr(num)

            elif state == ParseState.PS_ESCAPED_END:
                stack.pop()
                state = stack.pop()
                if state == ParseState.PS_LEFT:
                    left += escaped
                else:
                    right += escaped
                escaped = ""
            elif state == ParseState.PS_ESCAPED:
                if char in ["0", "1", "2", "3", "4", "5", "6", "7"]:
                    stack.append(state)
                    state = ParseState.PS_OCTAL
                    num = 0
                elif char.upper() == "X":
                    stack.append(state)
                    state = ParseState.PS_HEX
                    num = 0
                    pos += 1

                elif char == "a":
                    escaped = "\a"
                    stack.append(state)
                    state = ParseState.PS_ESCAPED_END
                    pos += 1
                elif char == "b":
                    escaped = "\b"
                    stack.append(state)
                    state = ParseState.PS_ESCAPED_END
                    pos += 1
                elif char == "e":
                    escaped = chr(27)
                    stack.append(state)
                    state = ParseState.PS_ESCAPED_END
                    pos += 1
                elif char == "f":
                    escaped = "\f"
                    stack.append(state)
                    state = ParseState.PS_ESCAPED_END
                    pos += 1
                elif char == "n":
                    escaped = "\n"
                    stack.append(state)
                    state = ParseState.PS_ESCAPED_END
                    pos += 1
                elif char == "r":
                    escaped = "\r"
                    stack.append(state)
                    state = ParseState.PS_ESCAPED_END
                    pos += 1
                elif char == "t":
                    escaped = "\t"
                    stack.append(state)
                    state = ParseState.PS_ESCAPED_END
                    pos += 1
                elif char == "v":
                    escaped = "\v"
                    stack.append(state)
                    state = ParseState.PS_ESCAPED_END
                    pos += 1
                elif char == "?":
                    escaped = chr(127)
                    stack.append(state)
                    state = ParseState.PS_ESCAPED_END
                    pos += 1
                elif char == "_":
                    escaped = " "
                    stack.append(state)
                    state = ParseState.PS_ESCAPED_END
                    pos += 1
                elif char == chr(0):  # pragma: no cover
                    raise EnvironmentError("Cannot parse coloring scheme")
                else:
                    escaped = char
                    stack.append(state)
                    state = ParseState.PS_ESCAPED_END
                    pos += 1
            elif state == ParseState.PS_CARRET:
                if "@" <= char <= "~":
                    escaped = chr(ord(char) & 0o37)
                elif char == "?":
                    escaped = chr(127)
                else:
                    raise EnvironmentError("Cannot parse coloring scheme")
                stack.append(state)
                state = ParseState.PS_ESCAPED_END
                pos += 1

            elif state == ParseState.PS_LEFT:
                if char == "\\":
                    stack.append(state)
                    state = ParseState.PS_ESCAPED
                    pos += 1
                    escaped = ""
                elif char == "=":
                    right = ""
                    state = ParseState.PS_RIGHT
                    pos += 1
                elif char == "^":
                    stack.append(state)
                    state = ParseState.PS_CARRET
                    pos += 1
                    escaped = ""
                else:
                    left += char
                    pos = pos + 1
            elif state == ParseState.PS_RIGHT:
                if char == "\\":
                    stack.append(state)
                    state = ParseState.PS_ESCAPED
                    pos += 1
                    escaped = ""
                elif char == ":":
                    if right:
                        process(left, right)
                    state = ParseState.PS_START
                    pos += 1
                elif char == "^":
                    stack.append(state)
                    state = ParseState.PS_CARRET
                    pos += 1
                    escaped = ""
                else:
                    right += char
                    pos += 1

        if state == ParseState.PS_CARRET:
            raise EnvironmentError("Cannot parse coloring scheme")
        if state in [ParseState.PS_HEX, ParseState.PS_OCTAL]:
            escaped = chr(num)
            state = stack.pop()
        if state == ParseState.PS_ESCAPED:
            stack.append(ParseState.PS_ESCAPED)
            state = ParseState.PS_ESCAPED_END
        if state == ParseState.PS_ESCAPED_END:
            stack.pop()
            state = stack.pop()
            if state == ParseState.PS_RIGHT:  # pragma no branch
                right += escaped
        if state == ParseState.PS_RIGHT and len(right):
            process(left, right)

    def paint(self, label: str, type: FileStatusType) -> str:
        mapping = {
            FileStatusType.FILE: self.color_indicator[GnuIndicators.FILE],
            FileStatusType.DIRECTORY: self.color_indicator[GnuIndicators.DIR],
        }
        color = mapping[type]
        if not color:
            color = self.color_indicator[GnuIndicators.NORM]
        if type == FileStatusType.FILE:
            for pattern, value in self.color_ext_type.items():
                if fnmatch(label, pattern):
                    color = value
                    break
        if color:
            if self._underline:
                underline = (
                    self.color_indicator[GnuIndicators.LEFT]
                    + "4"
                    + self.color_indicator[GnuIndicators.RIGHT]
                )
            else:
                underline = ""
            return (
                self.color_indicator[GnuIndicators.LEFT]
                + color
                + self.color_indicator[GnuIndicators.RIGHT]
                + underline
                + label
                + self.color_indicator[GnuIndicators.LEFT]
                + self.color_indicator[GnuIndicators.RESET]
                + self.color_indicator[GnuIndicators.RIGHT]
            )
        if self._underline:
            return style(label, underline=self._underline)
        else:
            return label


class BSDAttributes(enum.Enum):
    DIRECTORY = 1
    LINK = 2
    SOCKET = 3
    PIPE = 4
    EXECUTABLE = 5
    BLOCK = 6
    CHARACTER = 7
    EXECUTABLE_SETUID = 8
    EXECUTABLE_SETGID = 9
    DIRECTORY_WRITABLE_OTHERS_WITH_STICKY = 10
    DIRECTORY_WRITABLE_OTHERS_WITHOUT_STICKY = 11


class BSDPainter(BasePainter):
    def __init__(self, lscolors: str, *, underline: bool = False):
        self._underline = underline
        self._parse_lscolors(lscolors)

    def _parse_lscolors(self, lscolors: str) -> None:
        parts = chunks(lscolors, 2)
        self._colors: Dict[BSDAttributes, str] = {}
        num = 0
        for attr in BSDAttributes:
            self._colors[attr] = parts[num]
            num += 1

    def paint(self, label: str, type: FileStatusType) -> str:
        color = ""
        if type == FileStatusType.DIRECTORY:
            color = self._colors[BSDAttributes.DIRECTORY]
        if color:
            char_to_color = {
                "a": "black",
                "b": "red",
                "c": "green",
                "d": "brown",
                "e": "blue",
                "f": "magenta",
                "g": "cyan",
                "h": "white",
            }
            bold = None
            fg = bg = None
            if color[0].lower() in char_to_color.keys():
                fg = char_to_color[color[0].lower()]
                if color[0].isupper():
                    bold = True
            if color[1] in char_to_color.keys():
                bg = char_to_color[color[1]]
            if self._underline:
                underline: Optional[bool] = True
            else:
                underline = None
            if fg or bg or bold or underline:
                return style(label, fg=fg, bg=bg, bold=bold, underline=underline)
        if self._underline:
            return style(label, underline=self._underline)
        return label


def get_painter(color: bool, *, quote: bool = False) -> BasePainter:
    if color:
        ls_colors = os.getenv("LS_COLORS")
        if ls_colors:
            return GnuPainter(ls_colors, underline=quote)
        lscolors = os.getenv("LSCOLORS")
        if lscolors:
            return BSDPainter(lscolors, underline=quote)
    if quote:
        return QuotedPainter()
    else:
        return NonePainter()


class BaseFilesFormatter:
    @abc.abstractmethod
    def __call__(
        self, files: Sequence[FileStatus]
    ) -> Iterator[str]:  # pragma: no cover
        pass


class LongFilesFormatter(BaseFilesFormatter):
    permissions_mapping = {Action.MANAGE: "m", Action.WRITE: "w", Action.READ: "r"}

    file_types_mapping = {FileStatusType.FILE: "-", FileStatusType.DIRECTORY: "d"}

    def __init__(self, human_readable: bool, color: bool):
        self.human_readable = human_readable
        self.painter = get_painter(color)

    def _columns_for_file(self, file: FileStatus) -> Sequence[str]:

        type = self.file_types_mapping[file.type]
        permission = self.permissions_mapping[file.permission]

        date = time.strftime(TIME_FORMAT, time.localtime(file.modification_time))

        if self.human_readable:
            size = format_size(file.size).rstrip("B")
        else:
            size = str(file.size)

        name = self.painter.paint(file.name, file.type)

        return [f"{type}{permission}", f"{size}", f"{date}", f"{name}"]

    def __call__(self, files: Sequence[FileStatus]) -> Iterator[str]:
        if not files:
            return
        table = [self._columns_for_file(file) for file in files]
        widths = [0 for _ in table[0]]
        for row in table:
            for x in range(len(row)):
                cell_width = len(unstyle(row[x]))
                if widths[x] < cell_width:
                    widths[x] = cell_width
        for row in table:
            line = []
            for x in range(len(row)):
                if x == len(row) - 1:
                    line.append(row[x])
                else:
                    line.append(row[x].rjust(widths[x]))
            yield " ".join(line)


class SimpleFilesFormatter(BaseFilesFormatter):
    def __init__(self, color: bool):
        self.painter = get_painter(color)

    def __call__(self, files: Sequence[FileStatus]) -> Iterator[str]:
        for file in files:
            yield self.painter.paint(file.name, file.type)


class VerticalColumnsFilesFormatter(BaseFilesFormatter):
    def __init__(self, width: int, color: bool):
        self.width = width
        self.painter = get_painter(color)

    def __call__(self, files: Sequence[FileStatus]) -> Iterator[str]:
        if not files:
            return
        items = [self.painter.paint(file.name, file.type) for file in files]
        widths = [len(unstyle(item)) for item in items]
        # let`s check how many columns we can use
        test_count = 1
        while True:
            test_columns = chunks(widths, ceil(len(items) / test_count))
            test_columns_widths = [max(column) for column in test_columns]
            test_total_width = sum(test_columns_widths) + 2 * (len(test_columns) - 1)
            if test_count == 1 or test_total_width <= self.width:
                count = test_count
                columns_widths = test_columns_widths
                if test_total_width == self.width:
                    break

            if test_total_width >= self.width or len(test_columns) == len(items):
                break
            test_count = test_count + 1

        rows = transpose(chunks(items, ceil(len(items) / count)))
        for row in rows:
            formatted_row = []
            for i in range(len(row)):
                formatted = row[i]
                if i < len(row) - 1:
                    formatted = formatted.ljust(columns_widths[i])
                formatted_row.append(formatted)
            yield "  ".join(formatted_row)


class FilesSorter(str, enum.Enum):
    NAME = "name"
    SIZE = "size"
    TIME = "time"

    def key(self) -> Any:
        field = None
        if self == self.NAME:
            field = "name"
        elif self == self.SIZE:
            field = "size"
        elif self == self.TIME:
            field = "modification_time"
        assert field
        return operator.attrgetter(field)


# progress indicator


class DeleteProgress(AbstractDeleteProgress):
    def __init__(self, root: Root) -> None:
        self.painter = get_painter(root.color, quote=True)

    def fmt_url(self, url: URL, type: FileStatusType) -> str:
        return self.painter.paint(str(url), type)

    def delete(self, data: StorageProgressDelete) -> None:
        url_label = self.fmt_url(
            data.uri,
            FileStatusType.DIRECTORY if data.is_dir else FileStatusType.FILE,
        )
        click.echo(f"Removed: {url_label}")


class BaseStorageProgress(AbstractRecursiveFileProgress):
    @abc.abstractmethod
    def begin(self, src: URL, dst: URL) -> None:  # pragma: no cover
        pass

    def end(self) -> None:  # pragma: no cover
        pass


def create_storage_progress(root: Root, show_progress: bool) -> BaseStorageProgress:
    if show_progress:
        return TTYProgress(root)
    else:
        return StreamProgress(root)


def format_url(url: URL) -> str:
    if url.scheme == "file":
        path = _extract_path(url)
        return str(path)
    else:
        return str(url)


class StreamProgress(BaseStorageProgress):
    def __init__(self, root: Root) -> None:
        self.painter = get_painter(root.color, quote=True)
        self.verbose = root.verbosity > 0

    def fmt_url(self, url: URL, type: FileStatusType) -> str:
        label = format_url(url)
        return self.painter.paint(label, type)

    def begin(self, src: URL, dst: URL) -> None:
        if self.verbose:
            src_label = self.fmt_url(src, FileStatusType.DIRECTORY)
            dst_label = self.fmt_url(dst, FileStatusType.DIRECTORY)
            click.echo(f"Copy {src_label} -> {dst_label}")

    def start(self, data: StorageProgressStart) -> None:
        pass

    def complete(self, data: StorageProgressComplete) -> None:
        if not self.verbose:
            return
        src = self.fmt_url(data.src, FileStatusType.FILE)
        dst = self.fmt_url(data.dst, FileStatusType.FILE)
        click.echo(f"{src} -> {dst}")

    def step(self, data: StorageProgressStep) -> None:
        pass

    def enter(self, data: StorageProgressEnterDir) -> None:
        if not self.verbose:
            return
        src = self.fmt_url(data.src, FileStatusType.FILE)
        dst = self.fmt_url(data.dst, FileStatusType.FILE)
        click.echo(f"{src} -> {dst}")

    def leave(self, data: StorageProgressLeaveDir) -> None:
        pass

    def fail(self, data: StorageProgressFail) -> None:
        src = self.fmt_url(data.src, FileStatusType.FILE)
        dst = self.fmt_url(data.dst, FileStatusType.FILE)
        click.echo(
            click.style("Failure:", fg="red") + f" {src} -> {dst} [{data.message}]",
            err=True,
        )


class TTYProgress(BaseStorageProgress):
    HEIGHT = 25
    FLUSH_INTERVAL = 0.2
    time_factory = staticmethod(monotonic)

    def __init__(self, root: Root) -> None:
        self.painter = get_painter(root.color, quote=True)
        self.printer = TTYPrinter()
        self.half_width = (root.terminal_size[0] - 10) // 2
        self.full_width = root.terminal_size[0] - 20
        self.lines: List[Tuple[URL, bool, str]] = []
        self.cur_dir: Optional[URL] = None
        self.verbose = root.verbosity > 0
        self.first_line = self.last_line = 0
        self.last_update_time = self.time_factory()

    def fmt_url(self, url: URL, type: FileStatusType, *, half: bool) -> str:
        label = str(url)
        if half:
            width = self.half_width
        else:
            width = self.full_width
        while len(label) > width:
            parts = list(url.parts)
            if len(parts) < 2:
                break
            if parts[0] == "/":
                if len(parts) < 3:
                    slash = "/"
                    break
                slash, first, second, *last = parts
                if first == "...":
                    if last:
                        parts = ["..."] + last
                    else:
                        break
                else:
                    parts = ["...", second] + last
            else:
                slash = ""
                # len(parts) > 1 always
                first, second, *last = parts
                if first == "...":
                    if last:
                        parts = ["..."] + last
                    else:
                        break
                else:
                    parts = ["...", second] + last
            if url.host or slash:
                pre = f"//{url.host or ''}{slash}"
            else:
                pre = ""
            url = URL(f"{url.scheme}:{pre}{'/'.join(parts)}")
            label = str(url)
        return self.fmt_str(label, type)

    def fmt_str(self, label: str, type: FileStatusType) -> str:
        return self.painter.paint(label, type)

    def fmt_size(self, size: int) -> str:
        return format_size(size)

    def begin(self, src: URL, dst: URL) -> None:
        if self.verbose:
            click.echo("Copy")
            click.echo(self.fmt_str(str(src), FileStatusType.DIRECTORY))
            click.echo("=>")
            click.echo(self.fmt_str(str(dst), FileStatusType.DIRECTORY))
        else:
            src_label = self.fmt_url(src, FileStatusType.DIRECTORY, half=True)
            dst_label = self.fmt_url(dst, FileStatusType.DIRECTORY, half=True)
            click.echo(f"Copy {src_label} => {dst_label}")

    def end(self) -> None:
        self.flush()

    def enter(self, data: StorageProgressEnterDir) -> None:
        self._enter_dir(data.src)

    def _enter_dir(self, src: URL) -> None:
        fmt_src = self.fmt_url(src, FileStatusType.DIRECTORY, half=False)
        self.append(src, f"{fmt_src} ...", is_dir=True)
        self.cur_dir = src

    def leave(self, data: StorageProgressLeaveDir) -> None:
        src = self.fmt_url(data.src, FileStatusType.DIRECTORY, half=False)
        if self.lines and self.lines[-1][0] == data.src:
            self.lines.pop()
        self.append(data.src, f"{src} DONE", is_dir=True)
        self.flush()
        self.cur_dir = None

    def start(self, data: StorageProgressStart) -> None:
        src = self.fmt_str(data.src.name, FileStatusType.FILE)
        progress = 0
        current = self.fmt_size(0)
        total = self.fmt_size(data.size)
        self._append_file(data.src, f"{src} [{progress:.2f}%] {current} of {total}")

    def complete(self, data: StorageProgressComplete) -> None:
        src = self.fmt_str(data.src.name, FileStatusType.FILE)
        total = self.fmt_size(data.size)
        self.replace(data.src, f"{src} {total}")

    def step(self, data: StorageProgressStep) -> None:
        src = self.fmt_str(data.src.name, FileStatusType.FILE)
        progress = 100 * data.current / data.size if data.size else 100
        current = self.fmt_size(data.current)
        total = self.fmt_size(data.size)
        self.replace(data.src, f"{src} [{progress:.2f}%] {current} of {total}")

    def fail(self, data: StorageProgressFail) -> None:
        self.flush()
        src = self.fmt_str(str(data.src), FileStatusType.FILE)
        dst = self.fmt_str(str(data.dst), FileStatusType.FILE)
        click.echo(
            click.style("Failure:", fg="red") + f" {src} -> {dst} [{data.message}]",
            err=True,
        )
        # clear lines to sync with writing to stderr
        self.lines = []
        self.last_line = 0

    def _append_file(self, key: URL, msg: str) -> None:
        parent = key.parent
        if self.cur_dir is not None and self.cur_dir != parent:
            self._enter_dir(parent)
        self.append(key, msg)

    def append(self, key: URL, msg: str, is_dir: bool = False) -> None:
        self.lines.append((key, is_dir, msg))
        if len(self.lines) > self.HEIGHT:
            if not self.lines[0][1]:
                # top line is not a dir, drop it.
                del self.lines[0]
            elif self.lines[1][1]:
                # second line is a dir, drop the first line.
                del self.lines[0]
            else:
                # there is a file line under a dir line, drop the file line.
                del self.lines[1]
            self.first_line = 0
        self.last_line = len(self.lines)
        self.maybe_flush()

    def replace(self, key: URL, msg: str) -> None:
        for i in range(len(self.lines))[::-1]:
            line = self.lines[i]
            if line[0] == key:
                self.lines[i] = (key, False, msg)
                self.first_line = min(self.first_line, i)
                self.last_line = max(self.last_line, i + 1)
                self.maybe_flush()
                break
        else:
            self._append_file(key, msg)

    def maybe_flush(self) -> None:
        if (
            len(self.lines) < self.HEIGHT
            or self.time_factory() >= self.last_update_time + self.FLUSH_INTERVAL
        ):
            self.flush()

    def flush(self) -> None:
        text = "\n".join(
            msg for _, _, msg in self.lines[self.first_line : self.last_line]
        )
        self.printer.print(text, self.first_line)
        self.first_line = len(self.lines)
        self.last_line = 0
        self.last_update_time = self.time_factory()


@dataclass(frozen=True)
class Tree:
    name: str
    size: int
    folders: Sequence["Tree"]
    files: Sequence[FileStatus]


class TreeFormatter:
    ANSI_DELIMS = ["├", "└", "─", "│"]
    SIMPLE_DELIMS = ["+", "+", "-", "|"]

    def __init__(
        self, *, color: bool, size: bool, human_readable: bool, sort: str
    ) -> None:
        self._ident: List[bool] = []
        self._numdirs = 0
        self._numfiles = 0
        self._painter = get_painter(color, quote=True)
        if sys.platform != "win32":
            self._delims = self.ANSI_DELIMS
        else:
            self._delims = self.SIMPLE_DELIMS
        if human_readable:
            self._size_func = self._human_readable
        elif size:
            self._size_func = self._size
        else:
            self._size_func = self._none
        self._key = FilesSorter(sort).key()

    def __call__(self, tree: Tree) -> List[str]:
        ret = self.listdir(tree)
        ret.append("")
        ret.append(f"{self._numdirs} directories, {self._numfiles} files")
        return ret

    def listdir(self, tree: Tree) -> List[str]:
        ret = []
        items = sorted(tree.folders + tree.files, key=self._key)  # type: ignore
        ret.append(
            self.pre()
            + self._size_func(tree.size)
            + self._painter.paint(tree.name, FileStatusType.DIRECTORY)
        )
        for num, item in enumerate(items):
            if isinstance(item, Tree):
                self._numdirs += 1
                with self.ident(num == len(items) - 1):
                    ret.extend(self.listdir(item))
            else:
                self._numfiles += 1
                with self.ident(num == len(items) - 1):
                    ret.append(
                        self.pre()
                        + self._size_func(item.size)
                        + self._painter.paint(item.name, FileStatusType.FILE)
                    )
        return ret

    def pre(self) -> str:
        ret = []
        for last in self._ident[:-1]:
            if last:
                ret.append(" " * 4)
            else:
                ret.append(self._delims[3] + " " * 3)
        if self._ident:
            last = self._ident[-1]
            ret.append(self._delims[1] if last else self._delims[0])
            ret.append(self._delims[2] * 2)
            ret.append(" ")
        return "".join(ret)

    @contextlib.contextmanager
    def ident(self, last: bool) -> Iterator[None]:
        self._ident.append(last)
        try:
            yield
        finally:
            self._ident.pop()

    def _size(self, size: int) -> str:
        return f"[{size:>11}]  "

    def _human_readable(self, size: int) -> str:
        return f"[{format_size(size):>7}]  "

    def _none(self, size: int) -> str:
        return ""



import sys
from json import dumps

from neuromation.api import get

from .asyncio_utils import run
from .const import EX_DATAERR, EX_UNAVAILABLE, EX_USAGE


def error(message: str, exit_code: int) -> None:
    print(message)
    exit(exit_code)


async def async_main(action: str) -> None:
    if action == "store":
        error("Please use `neuro login` instead `docker login ...`", EX_UNAVAILABLE)
    elif action == "erase":
        print("Please use `neuro logout` instead `docker logout ...`", EX_UNAVAILABLE)
    else:
        async with get() as client:
            config = client.config
            registry = sys.stdin.readline().strip()
            neuro_registry = config.registry_url.host
            if registry != neuro_registry:
                error(
                    f"Unknown registry {registry}. "
                    "neuro configured with {neuro_registry}.",
                    EX_DATAERR,
                )
            token = await config.token()
            payload = {"Username": "token", "Secret": token}
            print(dumps(payload))


def main() -> None:
    if len(sys.argv) == 1 or sys.argv[1] not in ["store", "get", "erase"]:
        error(
            "Neuro Platform docker credential helper.\n:Service tool, not for use",
            EX_USAGE,
        )
    action = sys.argv[1]
    run(async_main(action))



import asyncio
import logging
from collections import namedtuple
from pathlib import Path
from typing import Any, AsyncIterator, Callable, List

import pytest
from yarl import URL

import neuromation
from neuromation.api import Cluster, Factory, Preset
from neuromation.api.config import _AuthConfig, _AuthToken, _ConfigData
from neuromation.cli import main
from neuromation.cli.const import EX_OK
from neuromation.cli.root import Root


SysCapWithCode = namedtuple("SysCapWithCode", ["out", "err", "code"])
log = logging.getLogger(__name__)


@pytest.fixture()
def nmrc_path(tmp_path: Path, token: str, auth_config: _AuthConfig) -> Path:
    nmrc_path = tmp_path / "conftest.nmrc"
    cluster_config = Cluster(
        registry_url=URL("https://registry-dev.neu.ro"),
        storage_url=URL("https://storage-dev.neu.ro"),
        users_url=URL("https://users-dev.neu.ro"),
        monitoring_url=URL("https://monitoring-dev.neu.ro"),
        secrets_url=URL("https://secrets-dev.neu.ro"),
        presets={
            "gpu-small": Preset(
                cpu=7, memory_mb=30 * 1024, gpu=1, gpu_model="nvidia-tesla-k80"
            ),
            "gpu-large": Preset(
                cpu=7, memory_mb=60 * 1024, gpu=1, gpu_model="nvidia-tesla-v100"
            ),
            "cpu-small": Preset(cpu=7, memory_mb=2 * 1024),
            "cpu-large": Preset(cpu=7, memory_mb=14 * 1024),
        },
        name="default",
    )
    config = _ConfigData(
        auth_config=auth_config,
        auth_token=_AuthToken.create_non_expiring(token),
        url=URL("https://dev.neu.ro/api/v1"),
        version=neuromation.__version__,
        cluster_name="default",
        clusters={cluster_config.name: cluster_config},
    )
    Factory(nmrc_path)._save(config)
    return nmrc_path


def create_root(config_path: Path) -> Root:
    return Root(
        color=False,
        tty=False,
        terminal_size=(80, 24),
        disable_pypi_version_check=True,
        network_timeout=60,
        config_path=config_path,
        verbosity=0,
        trace=False,
        trace_hide_token=True,
        command_path="",
        command_params=[],
        skip_gmp_stats=True,
        show_traceback=False,
    )


@pytest.fixture()
async def root(nmrc_path: Path, loop: asyncio.AbstractEventLoop) -> AsyncIterator[Root]:
    root = create_root(config_path=nmrc_path)
    await root.init_client()
    yield root
    await root.client.close()


@pytest.fixture()
async def root_no_logged_in(
    tmp_path: Path, loop: asyncio.AbstractEventLoop
) -> AsyncIterator[Root]:
    root = create_root(config_path=tmp_path)
    assert root._client is None
    yield root
    assert root._client is None


@pytest.fixture()
def run_cli(
    nmrc_path: Path, capfd: Any, tmp_path: Path
) -> Callable[[List[str]], SysCapWithCode]:
    def _run_cli(arguments: List[str]) -> SysCapWithCode:
        log.info("Run 'neuro %s'", " ".join(arguments))
        capfd.readouterr()

        code = EX_OK
        try:
            default_args = [
                "--show-traceback",
                "--disable-pypi-version-check",
                "--color=no",
            ]
            if "--neuromation-config" not in arguments:
                for arg in arguments:
                    if arg.startswith("--neuromation-config="):
                        break
                else:
                    default_args.append(f"--neuromation-config={nmrc_path}")

            main(default_args + arguments)
        except SystemExit as e:
            code = e.code
            pass
        out, err = capfd.readouterr()
        return SysCapWithCode(out.strip(), err.strip(), code)

    return _run_cli


@pytest.fixture()
def click_tty_emulation(monkeypatch: Any) -> None:
    monkeypatch.setattr("click._compat.isatty", lambda stream: True)



from textwrap import dedent

from click.testing import CliRunner

from neuromation.cli.main import MainGroup
from neuromation.cli.root import Root
from neuromation.cli.utils import DeprecatedGroup, command, group


def test_print() -> None:
    @group()
    def sub_command() -> None:
        pass

    @command()
    async def plain_cmd(root: Root) -> None:
        pass

    @group(cls=MainGroup)
    def main() -> None:
        pass

    main.add_command(sub_command)
    main.add_command(plain_cmd)
    main.skip_init = True

    runner = CliRunner()
    result = runner.invoke(main, [])
    assert result.exit_code == 0
    assert result.output == dedent(
        """\
        Usage: main [OPTIONS] COMMAND [ARGS]...

        Commands:
          sub-command

        Command Shortcuts:
          plain-cmd

        Use "main help <command>" for more information about a given command or topic.
        Use "main --options" for a list of global command-line options (applies to all
        commands).
    """
    )


def test_print_use_group_helpers() -> None:
    @group(cls=MainGroup)
    def main() -> None:
        pass

    @main.group()
    def sub_command() -> None:
        pass

    @main.command()
    async def plain_cmd(root: Root) -> None:
        pass

    main.skip_init = True
    runner = CliRunner()
    result = runner.invoke(main, [])
    assert result.exit_code == 0
    assert result.output == dedent(
        """\
        Usage: main [OPTIONS] COMMAND [ARGS]...

        Commands:
          sub-command

        Command Shortcuts:
          plain-cmd

        Use "main help <command>" for more information about a given command or topic.
        Use "main --options" for a list of global command-line options (applies to all
        commands).
    """
    )


def test_print_hidden() -> None:
    @group()
    def sub_command() -> None:
        pass

    @command(hidden=True)
    async def plain_cmd(root: Root) -> None:
        pass

    @group()
    def main() -> None:
        pass

    main.add_command(sub_command)
    main.add_command(plain_cmd)

    runner = CliRunner()
    result = runner.invoke(main, [])
    assert result.exit_code == 0
    assert result.output == dedent(
        """\
        Usage: main [OPTIONS] COMMAND [ARGS]...

        Commands:
          sub-command
    """
    )


def test_print_deprecated_group() -> None:
    @group()
    def sub_command() -> None:
        """
        Sub-command.
        """

    @group()
    def main() -> None:
        pass

    main.add_command(sub_command)
    main.add_command(DeprecatedGroup(sub_command, name="alias"))

    runner = CliRunner()
    result = runner.invoke(main, [])
    assert result.exit_code == 0
    assert result.output == dedent(
        """\
        Usage: main [OPTIONS] COMMAND [ARGS]...

        Commands:
          sub-command  Sub-command
          alias        Alias for sub-command
    """
    )


def test_print_deprecated_group_content() -> None:
    @group()
    def sub_command() -> None:
        """
        Sub-command.
        """

    @sub_command.command()
    async def cmd(root: Root) -> None:
        """Command.

        Detailed description is here.
        """

    @group(cls=MainGroup)
    def main() -> None:
        pass

    main.add_command(sub_command)
    main.add_command(DeprecatedGroup(sub_command, name="alias"))
    main.skip_init = True

    runner = CliRunner()
    result = runner.invoke(main, ["alias"])
    assert result.exit_code == 0
    assert result.output == dedent(
        """\
        Usage: main alias [OPTIONS] COMMAND [ARGS]...

          Alias for sub-command (DEPRECATED)

        Commands:
          cmd  Command
    """
    )


def test_print_deprecated_no_help() -> None:
    @command(deprecated=True)
    async def main(root: Root) -> None:
        pass

    runner = CliRunner()
    result = runner.invoke(main, ["--help"])
    assert result.exit_code == 0
    assert result.output == dedent(
        """\
        Usage: main [OPTIONS]

           (DEPRECATED)

        Options:
          --help  Show this message and exit.
    """
    )


def test_print_deprecated_with_help() -> None:
    @command(deprecated=True)
    async def main(root: Root) -> None:
        """Main help."""

    runner = CliRunner()
    result = runner.invoke(main, ["--help"])
    assert result.exit_code == 0
    assert result.output == dedent(
        """\
        Usage: main [OPTIONS]

          Main help. (DEPRECATED)

        Options:
          --help  Show this message and exit.
    """
    )


def test_print_help_with_examples() -> None:
    @command()
    async def main(root: Root) -> None:
        """
        Main help.

        Examples:

        # comment
        example

        """

    runner = CliRunner()
    result = runner.invoke(main, ["--help"])
    assert result.exit_code == 0
    assert result.output == dedent(
        """\
        Usage: main [OPTIONS]

          Main help.

        Examples:
          # comment
          example

        Options:
          --help  Show this message and exit.
    """
    )



import errno
import os
import subprocess
import sys
import textwrap
from pathlib import Path, PurePath
from typing import Tuple

import pytest
from yarl import URL

from neuromation.cli.const import EX_OSFILE
from neuromation.cli.formatters.storage import TreeFormatter
from tests.e2e import Helper
from tests.e2e.utils import FILE_SIZE_B


_Data = Tuple[str, str]


@pytest.mark.e2e
def test_e2e_storage(data: Tuple[Path, str], tmp_path: Path, helper: Helper) -> None:
    srcfile, checksum = data

    # Create directory for the test
    helper.mkdir("folder", parents=True)

    # Upload local file
    helper.check_upload_file_to_storage("foo", "folder", str(srcfile))

    # Confirm file has been uploaded
    helper.check_file_exists_on_storage("foo", "folder", FILE_SIZE_B)

    # Download into local file and confirm checksum
    helper.check_file_on_storage_checksum(
        "foo", "folder", checksum, str(tmp_path), "bar"
    )

    # Rename file on the storage
    helper.check_rename_file_on_storage("foo", "folder", "bar", "folder")
    helper.check_file_exists_on_storage("bar", "folder", FILE_SIZE_B)

    # Rename directory on the storage
    helper.check_rename_directory_on_storage("folder", "folder2")
    helper.check_file_exists_on_storage("bar", "folder2", FILE_SIZE_B)

    # Non-recursive removing should not have any effect
    with pytest.raises(IsADirectoryError, match=".+Target is a directory") as cm:
        helper.rm("folder2", recursive=False)
    assert cm.value.errno == errno.EISDIR
    helper.check_file_exists_on_storage("bar", "folder2", FILE_SIZE_B)


@pytest.mark.e2e
def test_empty_directory_ls_output(helper: Helper) -> None:
    helper.mkdir("")
    # Ensure output of ls - empty directory shall print nothing.
    captured = helper.run_cli(["storage", "ls", helper.tmpstorage])
    assert not captured.out


@pytest.mark.e2e
def test_ls_directory_itself(helper: Helper) -> None:
    helper.mkdir("")
    captured = helper.run_cli(["storage", "ls", "--directory", helper.tmpstorage])
    _, _, name = helper.tmpstoragename.rpartition("/")
    assert captured.out.splitlines() == [name]


@pytest.mark.e2e
def test_e2e_mkdir(helper: Helper) -> None:
    helper.run_cli(["storage", "mkdir", "--parents", helper.tmpstorage + "folder"])
    helper.check_dir_exists_on_storage("folder", "")

    # Create existing directory
    with pytest.raises(subprocess.CalledProcessError) as cm:
        helper.run_cli(["storage", "mkdir", helper.tmpstorage + "folder"])
    assert cm.value.returncode == EX_OSFILE
    helper.mkdir("folder", exist_ok=True)

    # Create a subdirectory in existing directory
    helper.run_cli(["storage", "mkdir", helper.tmpstorage + "folder/subfolder"])
    helper.check_dir_exists_on_storage("subfolder", "folder")

    # Create a subdirectory in non-existing directory
    with pytest.raises(subprocess.CalledProcessError) as cm:
        helper.run_cli(["storage", "mkdir", helper.tmpstorage + "parent/child"])
    assert cm.value.returncode == EX_OSFILE
    helper.check_dir_absent_on_storage("parent", "")
    helper.run_cli(
        ["storage", "mkdir", "--parents", helper.tmpstorage + "parent/child"]
    )
    helper.check_dir_exists_on_storage("parent", "")
    helper.check_dir_exists_on_storage("child", "parent")


@pytest.mark.e2e
def test_copy_local_file_to_platform_directory(helper: Helper, data2: _Data) -> None:
    srcfile, checksum = data2
    file_name = str(PurePath(srcfile).name)

    helper.mkdir("folder", parents=True)
    # Upload local file to existing directory
    helper.run_cli(["storage", "cp", srcfile, helper.tmpstorage + "/folder"])

    # Ensure file is there
    helper.check_file_exists_on_storage(file_name, "folder", FILE_SIZE_B // 3)


@pytest.mark.e2e
def test_copy_local_file_to_platform_directory_explicit(
    helper: Helper, data2: _Data
) -> None:
    srcfile, checksum = data2
    file_name = str(PurePath(srcfile).name)

    helper.mkdir("folder", parents=True)
    # Upload local file to existing directory
    helper.run_cli(["storage", "cp", "-t", helper.tmpstorage + "/folder", srcfile])

    # Ensure file is there
    helper.check_file_exists_on_storage(file_name, "folder", FILE_SIZE_B // 3)


@pytest.mark.e2e
def test_copy_local_single_file_to_platform_file(helper: Helper, data: _Data) -> None:
    # case when copy happens with rename to 'different_name.txt'
    srcfile, checksum = data
    file_name = str(PurePath(srcfile).name)

    helper.mkdir("folder", parents=True)
    # Upload local file to platform
    helper.run_cli(
        ["storage", "cp", srcfile, helper.tmpstorage + "/folder/different_name.txt"]
    )

    # Ensure file is there
    helper.check_file_exists_on_storage("different_name.txt", "folder", FILE_SIZE_B)
    helper.check_file_absent_on_storage(file_name, "folder")


@pytest.mark.e2e
def test_copy_local_single_file_to_platform_file_explicit(
    helper: Helper, data2: _Data
) -> None:
    # case when copy happens with rename to 'different_name.txt'
    srcfile, checksum = data2
    file_name = str(PurePath(srcfile).name)

    helper.mkdir("folder", parents=True)
    # Upload local file to platform
    helper.run_cli(
        [
            "storage",
            "cp",
            "-T",
            srcfile,
            helper.tmpstorage + "/folder/different_name.txt",
        ]
    )

    # Ensure file is there
    helper.check_file_exists_on_storage(
        "different_name.txt", "folder", FILE_SIZE_B // 3
    )
    helper.check_file_absent_on_storage(file_name, "folder")


@pytest.mark.e2e
def test_copy_local_to_platform_single_file_3(helper: Helper, data: _Data) -> None:
    # case when copy happens with rename to 'different_name.txt'
    srcfile, checksum = data

    helper.mkdir("")

    # Upload local file to non existing directory
    with pytest.raises(subprocess.CalledProcessError, match=str(EX_OSFILE)):
        captured = helper.run_cli(
            ["storage", "cp", srcfile, helper.tmpstorage + "/non_existing_dir/"]
        )
        assert not captured.err
        assert captured.out == ""

    # Ensure dir is not created
    helper.check_dir_absent_on_storage("non_existing_dir", "")


@pytest.mark.e2e
def test_e2e_copy_non_existing_platform_to_non_existing_local(
    helper: Helper, tmp_path: Path
) -> None:
    # Try downloading non existing file
    with pytest.raises(subprocess.CalledProcessError, match=str(EX_OSFILE)):
        helper.run_cli(
            [
                "storage",
                "cp",
                helper.tmpstorage + "/not-exist-foo",
                str(tmp_path / "not-exist-bar"),
            ]
        )


@pytest.mark.e2e
def test_e2e_copy_non_existing_platform_to_____existing_local(
    helper: Helper, tmp_path: Path
) -> None:
    # Try downloading non existing file
    with pytest.raises(subprocess.CalledProcessError, match=str(EX_OSFILE)):
        helper.run_cli(["storage", "cp", helper.tmpstorage + "/foo", str(tmp_path)])


@pytest.mark.e2e
def test_e2e_copy_no_sources_no_destination(helper: Helper) -> None:
    with pytest.raises(subprocess.CalledProcessError) as cm:
        helper.run_cli(["storage", "cp"])
    assert 'Missing argument "DESTINATION"' in cm.value.stderr


@pytest.mark.e2e
def test_e2e_copy_no_sources(helper: Helper) -> None:
    with pytest.raises(subprocess.CalledProcessError) as cm:
        helper.run_cli(["storage", "cp", helper.tmpstorage])
    assert 'Missing argument "SOURCES..."' in cm.value.stderr


@pytest.mark.e2e
def test_e2e_copy_no_sources_target_directory(helper: Helper) -> None:
    with pytest.raises(subprocess.CalledProcessError) as cm:
        helper.run_cli(["storage", "cp", "-t", helper.tmpstorage])
    assert 'Missing argument "SOURCES..."' in cm.value.stderr


@pytest.mark.e2e
def test_e2e_copy_target_directory_no_target_directory(
    helper: Helper, tmp_path: Path
) -> None:
    with pytest.raises(subprocess.CalledProcessError) as cm:
        helper.run_cli(["storage", "cp", "-t", helper.tmpstorage, "-T", str(tmp_path)])
    assert "Cannot combine" in cm.value.stderr


@pytest.mark.e2e
def test_e2e_copy_no_target_directory_extra_operand(
    helper: Helper, tmp_path: Path
) -> None:
    with pytest.raises(subprocess.CalledProcessError) as cm:
        helper.run_cli(
            ["storage", "cp", "-T", str(tmp_path), helper.tmpstorage, str(tmp_path)]
        )
    assert "Extra operand after " in cm.value.stderr


@pytest.mark.e2e
def test_copy_and_remove_multiple_files(
    helper: Helper, data2: _Data, data3: _Data, tmp_path: Path
) -> None:
    helper.mkdir("")
    # case when copy happens with rename to 'different_name.txt'
    srcfile, checksum = data2
    srcfile2, checksum2 = data3
    srcname = os.path.basename(srcfile)
    srcname2 = os.path.basename(srcfile2)

    # Upload local files
    captured = helper.run_cli(["storage", "cp", srcfile, srcfile2, helper.tmpstorage])
    assert captured.out == ""

    # Confirm files has been uploaded
    helper.check_file_exists_on_storage(srcname, "", FILE_SIZE_B // 3)
    helper.check_file_exists_on_storage(srcname2, "", FILE_SIZE_B // 5)

    # Download into local directory and confirm checksum
    targetdir = tmp_path / "bar"
    targetdir.mkdir()
    helper.run_cli(
        [
            "storage",
            "cp",
            f"{helper.tmpstorage}/{srcname}",
            f"{helper.tmpstorage}/{srcname2}",
            str(targetdir),
        ]
    )
    assert helper.hash_hex(targetdir / srcname) == checksum
    assert helper.hash_hex(targetdir / srcname2) == checksum2

    # Remove the files from platform
    captured = helper.run_cli(
        [
            "storage",
            "rm",
            f"{helper.tmpstorage}/{srcname}",
            f"{helper.tmpstorage}/{srcname2}",
        ]
    )
    assert captured.out == ""

    # Ensure files are not there
    helper.check_file_absent_on_storage(srcname, "")
    helper.check_file_absent_on_storage(srcname2, "")


@pytest.mark.e2e
def test_e2e_copy_recursive_to_platform(
    helper: Helper, nested_data: Tuple[str, str, str], tmp_path: Path
) -> None:
    helper.mkdir("")
    srcfile, checksum, dir_path = nested_data
    target_file_name = Path(srcfile).name

    # Upload local file
    captured = helper.run_cli(["storage", "cp", "-r", dir_path, helper.tmpstorage])
    # stderr has logs like "Using path ..."
    # assert not captured.err
    assert not captured.out

    helper.check_file_exists_on_storage(
        target_file_name, f"nested/directory/for/test", FILE_SIZE_B // 3
    )

    # Download into local directory and confirm checksum
    targetdir = tmp_path / "bar"
    targetdir.mkdir()
    helper.run_cli(["storage", "cp", "-r", f"{helper.tmpstorage}", str(targetdir)])
    targetfile = targetdir / "nested" / "directory" / "for" / "test" / target_file_name
    print("source file", srcfile)
    print("target file", targetfile)
    assert helper.hash_hex(targetfile) == checksum


@pytest.mark.e2e
def test_e2e_copy_recursive_file(helper: Helper, tmp_path: Path) -> None:
    helper.mkdir("")
    srcfile = tmp_path / "testfile"
    dstfile = tmp_path / "copyfile"
    srcfile.write_bytes(b"abc")

    captured = helper.run_cli(["storage", "cp", "-r", str(srcfile), helper.tmpstorage])
    assert not captured.out

    captured = helper.run_cli(
        ["storage", "cp", "-r", helper.tmpstorage + "testfile", str(dstfile)]
    )
    assert not captured.out

    assert dstfile.read_bytes() == b"abc"


@pytest.mark.e2e
def test_e2e_rename(helper: Helper) -> None:
    helper.mkdir("folder", parents=True)
    helper.run_cli(
        [
            "storage",
            "mv",
            helper.tmpstorage + "folder",
            helper.tmpstorage + "otherfolder",
        ]
    )
    helper.check_dir_absent_on_storage("folder", "")
    helper.check_dir_exists_on_storage("otherfolder", "")


@pytest.mark.e2e
def test_e2e_move_to_directory(helper: Helper) -> None:
    helper.mkdir("folder", parents=True)
    helper.mkdir("otherfolder", parents=True)
    helper.run_cli(
        [
            "storage",
            "mv",
            helper.tmpstorage + "folder",
            helper.tmpstorage + "otherfolder",
        ]
    )
    helper.check_dir_absent_on_storage("folder", "")
    helper.check_dir_exists_on_storage("otherfolder", "")
    helper.check_dir_exists_on_storage("folder", "otherfolder")


@pytest.mark.e2e
def test_e2e_move_to_directory_explicitly(helper: Helper) -> None:
    helper.mkdir("folder", parents=True)
    helper.mkdir("otherfolder", parents=True)
    helper.run_cli(
        [
            "storage",
            "mv",
            "-t",
            helper.tmpstorage + "otherfolder",
            helper.tmpstorage + "folder",
        ]
    )
    helper.check_dir_absent_on_storage("folder", "")
    helper.check_dir_exists_on_storage("otherfolder", "")
    helper.check_dir_exists_on_storage("folder", "otherfolder")


@pytest.mark.e2e
def test_e2e_move_content_to_directory(helper: Helper) -> None:
    helper.mkdir("folder/subfolder", parents=True)
    helper.mkdir("otherfolder", parents=True)
    helper.run_cli(
        [
            "storage",
            "mv",
            "-T",
            helper.tmpstorage + "folder",
            helper.tmpstorage + "otherfolder",
        ]
    )
    helper.check_dir_absent_on_storage("folder", "")
    helper.check_dir_exists_on_storage("subfolder", "otherfolder")


@pytest.mark.e2e
def test_e2e_move_no_sources_no_destination(helper: Helper) -> None:
    with pytest.raises(subprocess.CalledProcessError) as cm:
        helper.run_cli(["storage", "mv"])
    assert 'Missing argument "DESTINATION"' in cm.value.stderr


@pytest.mark.e2e
def test_e2e_move_no_sources(helper: Helper) -> None:
    with pytest.raises(subprocess.CalledProcessError) as cm:
        helper.run_cli(["storage", "mv", helper.tmpstorage])
    assert 'Missing argument "SOURCES..."' in cm.value.stderr


@pytest.mark.e2e
def test_e2e_move_no_sources_target_directory(helper: Helper) -> None:
    with pytest.raises(subprocess.CalledProcessError) as cm:
        helper.run_cli(["storage", "mv", "-t", helper.tmpstorage])
    assert 'Missing argument "SOURCES..."' in cm.value.stderr


@pytest.mark.e2e
def test_e2e_move_target_directory_no_target_directory(helper: Helper) -> None:
    with pytest.raises(subprocess.CalledProcessError) as cm:
        helper.run_cli(
            [
                "storage",
                "mv",
                "-t",
                helper.tmpstorage + "/foo",
                "-T",
                helper.tmpstorage + "/bar",
            ]
        )
    assert "Cannot combine" in cm.value.stderr


@pytest.mark.e2e
def test_e2e_move_no_target_directory_extra_operand(helper: Helper) -> None:
    with pytest.raises(subprocess.CalledProcessError) as cm:
        helper.run_cli(
            [
                "storage",
                "mv",
                "-T",
                helper.tmpstorage + "/foo",
                helper.tmpstorage + "/bar",
                helper.tmpstorage + "/baz",
            ]
        )
    assert "Extra operand after " in cm.value.stderr


@pytest.mark.e2e
def test_e2e_glob(tmp_path: Path, helper: Helper) -> None:
    # Create files and directories and copy them to storage
    helper.mkdir("")
    folder = tmp_path / "folder"
    folder.mkdir()
    (folder / "subfolder").mkdir()
    (folder / "foo").write_bytes(b"foo")
    (folder / "bar").write_bytes(b"bar")
    (folder / "baz").write_bytes(b"baz")
    helper.run_cli(
        [
            "storage",
            "cp",
            "-r",
            tmp_path.as_uri() + "/f*",
            helper.tmpstorage + "/folder",
        ]
    )
    captured = helper.run_cli(["storage", "ls", helper.tmpstorage + "/folder"])
    assert sorted(captured.out.splitlines()) == ["bar", "baz", "foo", "subfolder"]

    # Move files with pattern
    helper.run_cli(
        [
            "storage",
            "mv",
            helper.tmpstorage + "/folder/[bf]*",
            helper.tmpstorage + "/folder/subfolder",
        ]
    )
    captured = helper.run_cli(
        ["storage", "ls", helper.tmpstorage + "/folder/subfolder"]
    )
    assert sorted(captured.out.splitlines()) == ["bar", "baz", "foo"]

    # Download files with pattern
    download = tmp_path / "download"
    download.mkdir()
    helper.run_cli(["storage", "cp", helper.tmpstorage + "/**/b*", str(download)])
    assert sorted(download.iterdir()) == [download / "bar", download / "baz"]

    # Remove files with pattern
    helper.run_cli(["storage", "rm", helper.tmpstorage + "/**/b*"])
    captured = helper.run_cli(
        ["storage", "ls", helper.tmpstorage + "/folder/subfolder"]
    )
    assert sorted(captured.out.splitlines()) == ["foo"]

    # Test subcommand "glob"
    captured = helper.run_cli(["storage", "glob", helper.tmpstorage + "/**"])
    prefix = (
        f"storage://{helper.cluster_name}/{helper.username}/"
        f"{URL(helper.tmpstorage).path}"
    )
    assert sorted(captured.out.splitlines()) == [
        prefix,
        prefix + "folder",
        prefix + "folder/subfolder",
        prefix + "folder/subfolder/foo",
    ]


@pytest.mark.e2e
def test_e2e_no_glob(tmp_path: Path, helper: Helper) -> None:
    # Create files and directories and copy them to storage
    helper.mkdir("")
    dir = tmp_path / "[d]"
    dir.mkdir()
    (dir / "f").write_bytes(b"f")
    (dir / "[df]").write_bytes(b"[df]")
    helper.run_cli(
        [
            "storage",
            "cp",
            "-r",
            "--no-glob",
            tmp_path.as_uri() + "/[d]",
            helper.tmpstorage + "/d",
        ]
    )

    # Move files with literal path
    helper.run_cli(
        ["storage", "mv", "--no-glob", helper.tmpstorage + "/d/[df]", helper.tmpstorage]
    )
    captured = helper.run_cli(["storage", "ls", helper.tmpstorage])
    assert sorted(captured.out.splitlines()) == ["[df]", "d"]

    # Download files with literal path
    download = tmp_path / "download"
    download.mkdir()
    helper.run_cli(
        ["storage", "cp", "--no-glob", helper.tmpstorage + "/[df]", str(download)]
    )
    assert sorted(download.iterdir()) == [download / "[df]"]

    # Remove files with literal path
    helper.run_cli(["storage", "rm", "--no-glob", helper.tmpstorage + "/[df]"])
    captured = helper.run_cli(["storage", "ls", helper.tmpstorage])
    assert sorted(captured.out.splitlines()) == ["d"]


@pytest.mark.e2e
def test_e2e_cp_filter(tmp_path: Path, helper: Helper) -> None:
    # Create files and directories and copy them to storage
    helper.mkdir("")
    folder = tmp_path / "folder"
    folder.mkdir()
    (folder / "subfolder").mkdir()
    (folder / "foo").write_bytes(b"foo")
    (folder / "bar").write_bytes(b"bar")
    (folder / "baz").write_bytes(b"baz")

    helper.run_cli(
        [
            "storage",
            "cp",
            "-r",
            "--exclude",
            "*",
            "--include",
            "b??",
            "--exclude",
            "*z",
            tmp_path.as_uri() + "/folder",
            helper.tmpstorage + "/filtered",
        ]
    )
    captured = helper.run_cli(["storage", "ls", helper.tmpstorage + "/filtered"])
    assert captured.out.splitlines() == ["bar"]

    # Copy all files to storage
    helper.run_cli(
        [
            "storage",
            "cp",
            "-r",
            tmp_path.as_uri() + "/folder",
            helper.tmpstorage + "/folder",
        ]
    )

    # Copy filtered files from storage
    helper.run_cli(
        [
            "storage",
            "cp",
            "-r",
            "--exclude",
            "*",
            "--include",
            "b??",
            "--exclude",
            "*z",
            helper.tmpstorage + "/folder",
            tmp_path.as_uri() + "/filtered",
        ]
    )
    assert os.listdir(tmp_path / "filtered") == ["bar"]


@pytest.mark.e2e
def test_e2e_ls_skip_hidden(tmp_path: Path, helper: Helper) -> None:
    # Create files and directories and copy them to storage
    helper.mkdir("")
    folder = tmp_path / "folder"
    folder.mkdir()
    (folder / "foo").write_bytes(b"foo")
    (folder / ".bar").write_bytes(b"bar")

    helper.run_cli(
        ["storage", "cp", "-r", tmp_path.as_uri() + "/folder", helper.tmpstorage]
    )

    captured = helper.run_cli(["storage", "ls", helper.tmpstorage + "/folder"])
    assert captured.out.splitlines() == ["foo"]


@pytest.mark.e2e
def test_e2e_ls_show_hidden(tmp_path: Path, helper: Helper) -> None:
    # Create files and directories and copy them to storage
    helper.mkdir("")
    folder = tmp_path / "folder"
    folder.mkdir()
    (folder / "foo").write_bytes(b"foo")
    (folder / ".bar").write_bytes(b"bar")

    helper.run_cli(
        ["storage", "cp", "-r", tmp_path.as_uri() + "/folder", helper.tmpstorage]
    )

    captured = helper.run_cli(["storage", "ls", "--all", helper.tmpstorage + "/folder"])
    assert captured.out.splitlines() == [".bar", "foo"]


@pytest.mark.e2e
def test_tree(helper: Helper, data: _Data, tmp_path: Path) -> None:
    folder = tmp_path / "folder"
    folder.mkdir()
    (folder / "foo").write_bytes(b"foo")
    (folder / "bar").write_bytes(b"bar")
    subfolder = folder / "folder"
    subfolder.mkdir()
    (subfolder / "baz").write_bytes(b"baz")

    helper.run_cli(["storage", "cp", "-r", folder.as_uri(), helper.tmpstorage])

    capture = helper.run_cli(["storage", "tree", helper.tmpstorage])
    assert capture.err == ""

    expected = textwrap.dedent(
        f"""\
         '{helper.tmpstorage}'
         ├── 'bar'
         ├── 'folder'
         │   └── 'baz'
         └── 'foo'

         1 directories, 3 files"""
    )
    if sys.platform == "win32":
        trans = str.maketrans(
            "".join(TreeFormatter.ANSI_DELIMS), "".join(TreeFormatter.SIMPLE_DELIMS)
        )
        expected = expected.translate(trans)
    assert capture.out == expected



JOB_GPU_NUMBER = 0
JOB_GPU_MODEL = "nvidia-tesla-k80"
JOB_CPU_NUMBER = 0.1
JOB_MEMORY_AMOUNT = "1G"
JOB_DEBUG_LOCAL_PORT = 31234
JOB_SSH_USER = "root"

GPU_MODELS = ("nvidia-tesla-k80", "nvidia-tesla-p4", "nvidia-tesla-v100")



from pathlib import Path

import aiohttp
import pytest

from neuromation.cli.root import Root


@pytest.fixture
def root_uninitialized() -> Root:
    return Root(
        color=False,
        tty=False,
        terminal_size=(80, 25),
        disable_pypi_version_check=False,
        network_timeout=60,
        config_path=Path("~/.neuro"),
        verbosity=0,
        trace=False,
        trace_hide_token=True,
        command_path="",
        command_params=[],
        skip_gmp_stats=True,
        show_traceback=False,
    )


def test_timeout(root_uninitialized: Root) -> None:
    assert root_uninitialized.timeout == aiohttp.ClientTimeout(None, None, 60, 60)


class TestTokenSanitization:
    @pytest.mark.parametrize("auth", ["Bearer", "Basic", "Digest", "Mutual"])
    def test_sanitize_header_value_single_token(
        self, root_uninitialized: Root, auth: str
    ) -> None:
        line = f"{auth} eyJhbGciOiJI.eyJzdW0NTY3.SfKxwRJ_SsM"
        expected = f"{auth} eyJhb<hidden 26 chars>J_SsM"
        line_safe = root_uninitialized._sanitize_header_value(line)
        assert line_safe == expected

    @pytest.mark.parametrize("auth", ["Bearer", "Basic", "Digest", "Mutual"])
    def test_sanitize_header_value_many_tokens(
        self, root_uninitialized: Root, auth: str
    ) -> None:
        num = 10
        line = f"{auth} eyJhbGcOiJI.eyJzdTY3.SfKxwRJ_SsM " * num
        expected = f"{auth} eyJhb<hidden 22 chars>J_SsM " * num
        line_safe = root_uninitialized._sanitize_header_value(line)
        assert line_safe == expected

    @pytest.mark.parametrize("auth", ["Bearer", "Basic", "Digest", "Mutual"])
    def test_sanitize_header_value_not_a_token(
        self, root_uninitialized: Root, auth: str
    ) -> None:
        line = f"{auth} not_a_jwt"
        line_safe = root_uninitialized._sanitize_header_value(line)
        assert line_safe == f"{auth} not_a_jwt"

    def test_sanitize_token_replaced_overall(self, root_uninitialized: Root) -> None:
        token = "a.b.c"
        line_safe = root_uninitialized._sanitize_token(token)
        assert line_safe == "<hidden 5 chars>"



import pytest

import neuromation
from tests.e2e import Helper


@pytest.mark.e2e
def test_print_version(helper: Helper) -> None:
    expected_out = f"Neuro Platform Client {neuromation.__version__}"

    captured = helper.run_cli(["--version"])
    assert not captured.err
    assert captured.out == expected_out


@pytest.mark.e2e
def test_print_options(helper: Helper) -> None:
    captured = helper.run_cli(["--options"])
    assert not captured.err
    assert "Options" in captured.out


@pytest.mark.e2e
def test_print_config(helper: Helper) -> None:
    captured = helper.run_cli(["config", "show"])
    assert not captured.err
    assert "API URL: https://dev.neu.ro/api/v1" in captured.out


@pytest.mark.e2e
def test_print_config_token(helper: Helper) -> None:
    captured = helper.run_cli(["config", "show-token"])
    assert not captured.err
    assert captured.out  # some secure information was printed


@pytest.mark.e2e
def test_print_get_clusters(helper: Helper) -> None:
    captured = helper.run_cli(["config", "get-clusters"])
    assert not captured.err
    assert "Name: default" in captured.out
    assert "Presets" in captured.out


@pytest.mark.e2e
def test_root_trace_hide_token_default_true(helper: Helper) -> None:
    captured = helper.run_cli(["--trace", "ls"])
    assert "Authorization: Bearer " in captured.err
    assert "<hidden " in captured.err
    assert " chars>" in captured.err



import os
import sys
from pathlib import Path
from typing import Any, Callable, Dict
from unittest import mock

import aiohttp
import pytest
from aiohttp import web
from aiohttp.test_utils import TestServer as _TestServer
from yarl import URL

import neuromation
import neuromation.api.config_factory
from neuromation.api import Cluster, ConfigError, Factory
from neuromation.api.config import _AuthConfig, _AuthToken, _ConfigData
from neuromation.api.errors import AuthException
from tests import _TestServerFactory


@pytest.fixture
def tmp_home(tmp_path: Path, monkeypatch: Any) -> Path:
    monkeypatch.setattr(Path, "home", lambda: tmp_path)  # Like as it's not enough
    if os.name == "nt" and sys.version_info >= (3, 8):
        monkeypatch.setenv("USERPROFILE", str(tmp_path))
    else:
        monkeypatch.setenv("HOME", str(tmp_path))

    return tmp_path


@pytest.fixture
def config_dir(
    tmp_home: Path, token: str, auth_config: _AuthConfig, cluster_config: Cluster
) -> Path:
    config_path = tmp_home / ".neuro"
    _create_config(config_path, token, auth_config, cluster_config)
    return config_path


def _create_config(
    nmrc_path: Path, token: str, auth_config: _AuthConfig, cluster_config: Cluster
) -> str:
    config = _ConfigData(
        auth_config=auth_config,
        auth_token=_AuthToken.create_non_expiring(token),
        url=URL("https://dev.neu.ro/api/v1"),
        version=neuromation.__version__,
        cluster_name=cluster_config.name,
        clusters={cluster_config.name: cluster_config},
    )
    Factory(nmrc_path)._save(config)
    assert nmrc_path.exists()
    return token


@pytest.fixture
async def mock_for_login(
    aiohttp_server: _TestServerFactory,
    token: str,
    aiohttp_unused_port: Callable[[], int],
) -> _TestServer:

    callback_urls = [
        f"http://127.0.0.1:{aiohttp_unused_port()}",
        f"http://127.0.0.1:{aiohttp_unused_port()}",
        f"http://127.0.0.1:{aiohttp_unused_port()}",
    ]

    async def config_handler(request: web.Request) -> web.Response:
        config_json: Dict[str, Any] = {
            "auth_url": str(srv.make_url("/authorize")),
            "token_url": str(srv.make_url("/oauth/token")),
            "logout_url": str(srv.make_url("/v2/logout")),
            "client_id": "banana",
            "audience": "https://test.dev.neuromation.io",
            "headless_callback_url": str(srv.make_url("/oauth/show-code")),
            "callback_urls": callback_urls,
            "success_redirect_url": "https://neu.ro/#test",
        }

        if (
            "Authorization" in request.headers
            and "incorrect" not in request.headers["Authorization"]
        ):
            config_json.update(
                {
                    "clusters": [
                        {
                            "name": "default",
                            "registry_url": "https://registry-dev.test.com",
                            "storage_url": "https://storage-dev.test.com",
                            "users_url": "https://users-dev.test.com",
                            "monitoring_url": "https://monitoring-dev.test.com",
                            "secrets_url": "https://secrets-dev.test.com",
                            "resource_presets": [
                                {
                                    "name": "gpu-small",
                                    "cpu": 7,
                                    "memory_mb": 30 * 1024,
                                    "gpu": 1,
                                    "gpu_model": "nvidia-tesla-k80",
                                },
                                {
                                    "name": "gpu-large",
                                    "cpu": 7,
                                    "memory_mb": 60 * 1024,
                                    "gpu": 1,
                                    "gpu_model": "nvidia-tesla-v100",
                                },
                                {"name": "cpu-small", "cpu": 2, "memory_mb": 2 * 1024},
                                {"name": "cpu-large", "cpu": 3, "memory_mb": 14 * 1024},
                            ],
                        }
                    ]
                }
            )
        return web.json_response(config_json)

    async def show_code(request: web.Request) -> web.Response:
        return web.json_response({})

    async def authorize(request: web.Request) -> web.Response:
        url = URL(request.query["redirect_uri"]).with_query(code="test_auth_code")
        raise web.HTTPSeeOther(location=url)

    async def new_token(request: web.Request) -> web.Response:
        return web.json_response(
            {"access_token": token, "expires_in": 3600, "refresh_token": token}
        )

    app = web.Application()
    app.router.add_get("/config", config_handler)
    app.router.add_get("/oauth/show-code", show_code)
    app.router.add_get("/authorize", authorize)
    app.router.add_post("/oauth/token", new_token)
    srv = await aiohttp_server(app)
    return srv


class TestConfigFileInteraction:
    async def test_config_file_absent(self, tmp_home: Path) -> None:
        with pytest.raises(ConfigError, match=r"file.+not exists"):
            await Factory().get()

    async def test_config_dir_is_file(self, tmp_home: Path) -> None:
        Path(tmp_home / ".neuro").write_text("something")
        with pytest.raises(ConfigError, match=r"not a directory"):
            await Factory().get()

    async def test_config_file_is_dir(self, tmp_home: Path) -> None:
        path = Path(tmp_home / ".neuro")
        path.mkdir()
        (path / "db").mkdir()
        with pytest.raises(ConfigError, match=r"not a regular file"):
            await Factory().get()

    async def test_default_path(
        self,
        tmp_home: Path,
        token: str,
        auth_config: _AuthConfig,
        cluster_config: Cluster,
    ) -> None:
        token = _create_config(tmp_home / ".neuro", token, auth_config, cluster_config)
        client = await Factory().get()
        await client.close()
        assert await client.config.token() == token

    async def test_preset_serialization(
        self,
        tmp_home: Path,
        token: str,
        auth_config: _AuthConfig,
        cluster_config: Cluster,
    ) -> None:
        _create_config(tmp_home / ".neuro", token, auth_config, cluster_config)
        client = await Factory().get()
        await client.close()
        assert len(client.presets) > 0
        assert not client.presets["cpu-large"].is_preemptible
        assert client.presets["cpu-large-p"].is_preemptible

    async def test_shorten_path(
        self,
        tmp_home: Path,
        token: str,
        auth_config: _AuthConfig,
        cluster_config: Cluster,
    ) -> None:
        token = _create_config(
            tmp_home / "test.nmrc", token, auth_config, cluster_config
        )
        client = await Factory(Path("~/test.nmrc")).get()
        await client.close()
        assert await client.config.token() == token

    async def test_full_path(
        self,
        tmp_home: Path,
        token: str,
        auth_config: _AuthConfig,
        cluster_config: Cluster,
    ) -> None:
        config_path = tmp_home / "test.nmrc"
        token = _create_config(config_path, token, auth_config, cluster_config)
        client = await Factory(config_path).get()
        await client.close()
        assert await client.config.token() == token

    @pytest.mark.skipif(
        sys.platform == "win32",
        reason="Windows does not supports UNIX-like permissions",
    )
    async def test_file_permissions(self, config_dir: Path) -> None:
        Path(config_dir).chmod(0o777)
        with pytest.raises(ConfigError, match=r"permission"):
            await Factory().get()


class TestLogin:
    async def show_dummy_browser(self, url: URL) -> None:
        async with aiohttp.ClientSession() as client:
            await client.get(url, allow_redirects=True)

    async def test_login_already_logged(self, config_dir: Path) -> None:
        with pytest.raises(ConfigError, match=r"already exists"):
            await Factory().login(self.show_dummy_browser)

    async def test_normal_login(
        self, tmp_home: Path, mock_for_login: _TestServer
    ) -> None:
        await Factory().login(self.show_dummy_browser, url=mock_for_login.make_url("/"))
        nmrc_path = tmp_home / ".neuro"
        assert Path(nmrc_path).exists(), "Config file not written after login "


class TestLoginWithToken:
    async def test_login_with_token_already_logged(self, config_dir: Path) -> None:
        with pytest.raises(ConfigError, match=r"already exists"):
            await Factory().login_with_token(token="tokenstr")

    async def test_normal_login(
        self, tmp_home: Path, mock_for_login: _TestServer
    ) -> None:
        await Factory().login_with_token(
            token="tokenstr", url=mock_for_login.make_url("/")
        )
        nmrc_path = tmp_home / ".neuro"
        assert Path(nmrc_path).exists(), "Config file not written after login "

    async def test_incorrect_token(
        self, tmp_home: Path, mock_for_login: _TestServer
    ) -> None:
        with pytest.raises(AuthException):
            await Factory().login_with_token(
                token="incorrect", url=mock_for_login.make_url("/")
            )
        nmrc_path = tmp_home / ".neuro"
        assert not Path(nmrc_path).exists(), "Config file not written after login "


class TestHeadlessLogin:
    async def test_login_headless_already_logged(self, config_dir: Path) -> None:
        async def get_auth_code_cb(url: URL) -> str:
            return ""

        with pytest.raises(ConfigError, match=r"already exists"):
            await Factory().login_headless(get_auth_code_cb)

    async def test_normal_login(
        self, tmp_home: Path, mock_for_login: _TestServer
    ) -> None:
        async def get_auth_code_cb(url: URL) -> str:
            assert url.with_query(None) == mock_for_login.make_url("/authorize")

            assert dict(url.query) == dict(
                response_type="code",
                code_challenge=mock.ANY,
                code_challenge_method="S256",
                client_id="banana",
                redirect_uri=str(mock_for_login.make_url("/oauth/show-code")),
                scope="offline_access",
                audience="https://test.dev.neuromation.io",
            )
            return "test_code"

        await Factory().login_headless(
            get_auth_code_cb, url=mock_for_login.make_url("/")
        )
        nmrc_path = tmp_home / ".neuro"
        assert Path(nmrc_path).exists(), "Config file not written after login "


class TestLogout:
    async def test_logout_no_browser_callback(self, config_dir: Path) -> None:
        await Factory().logout()
        assert not config_dir.exists(), "Config not removed after logout\n" + "\n".join(
            [p.name for p in config_dir.iterdir()]
        )

    async def test_logout_with_browser_callback(
        self, auth_config: _AuthConfig, config_dir: Path
    ) -> None:
        async def show_browser(url: URL) -> None:
            expected_logout_url = auth_config.logout_url.with_query(
                client_id=auth_config.client_id
            )
            assert url == expected_logout_url

        await Factory().logout(show_browser)
        assert not config_dir.exists(), "Config not removed after logout\n" + "\n".join(
            [p.name for p in config_dir.iterdir()]
        )



import re
from typing import Callable
from uuid import uuid4

import click
from yarl import URL

from neuromation.api import RemoteImage
from neuromation.api.url_utils import CLUSTER_SCHEMES


URIFormatter = Callable[[URL], str]
ImageFormatter = Callable[[RemoteImage], str]


BOLD = re.compile(r"(?P<mark>\*\*|__)(?P<content>\S.*?\S)(?P=mark)")
EMPHASIS = re.compile(r"(?P<mark>\*|_)(?P<content>\S.*?\S)(?P=mark)")
INLINE_CODE = re.compile(r"`(?P<content>.*?)`")
CODE_BLOCK = re.compile(r"```(?P<content>.*?)```", re.DOTALL | re.MULTILINE)


def uri_formatter(username: str, cluster_name: str) -> URIFormatter:
    def formatter(uri: URL) -> str:
        if uri.scheme in CLUSTER_SCHEMES:
            if uri.host == cluster_name:
                assert uri.path[0] == "/"
                path = uri.path.lstrip("/")
                owner, _, rest = path.partition("/")
                if owner == username:
                    return f"{uri.scheme}:{rest.lstrip('/')}"
                return f"{uri.scheme}:/{path}"
        return str(uri)

    return formatter


def image_formatter(uri_formatter: URIFormatter) -> ImageFormatter:
    def formatter(image: RemoteImage) -> str:
        image_str = str(image)
        if image_str.startswith("image://"):
            return uri_formatter(URL(image_str))
        else:
            return image_str

    return formatter


def apply_styling(txt: str) -> str:
    REPLACES = {}

    # code blocks
    match = CODE_BLOCK.search(txt)
    while match is not None:
        label = f"REPLACE-CODE-BLOCK-{uuid4()}"
        lines = []
        first = True
        for line in match.group("content").splitlines():
            if first and not line:
                first = False
                continue
            lines.append(click.style(line, dim=True))
        REPLACES[label] = "\n".join(lines)
        txt = txt[: match.start()] + label + txt[match.end() :]
        match = CODE_BLOCK.search(txt)

    # inline codes
    match = INLINE_CODE.search(txt)
    while match is not None:
        label = f"REPLACE-INLINE-CODE-{uuid4()}"
        REPLACES[label] = click.style(match.group("content"), bold=True, dim=True)
        txt = txt[: match.start()] + label + txt[match.end() :]
        match = INLINE_CODE.search(txt)

    txt = BOLD.sub(click.style(r"\g<content>", bold=True), txt)
    txt = EMPHASIS.sub(click.style(r"\g<content>", underline=True), txt)
    for key, value in REPLACES.items():
        txt = txt.replace(key, value)
    return txt



from typing import Any, Callable, Iterable, Optional, Type

import click
from click.utils import make_default_short_help

from .formatters.utils import apply_styling


class Command(click.Command):
    def format_help(self, ctx: click.Context, formatter: click.HelpFormatter) -> None:
        if self.help is None:
            return
        formatter.write_paragraph()
        formatter.write(apply_styling(self.help))

    def get_short_help_str(self, limit: int = 45) -> str:
        if self.help is None:
            return ""
        head, *tail = self.help.split("\n", 1)
        return make_default_short_help(head.strip(" *_."))


def command(
    name: Optional[str] = None, cls: Type[Command] = Command, **kwargs: Any
) -> Command:
    return click.command(name=name, cls=cls, **kwargs)  # type: ignore


class Group(click.Group):
    def command(
        self, *args: Any, **kwargs: Any
    ) -> Callable[[Callable[..., Any]], Command]:
        def decorator(f: Callable[..., Any]) -> Command:
            cmd = command(*args, **kwargs)(f)
            self.add_command(cmd)
            return cmd

        return decorator

    def group(
        self, *args: Any, **kwargs: Any
    ) -> Callable[[Callable[..., Any]], "Group"]:
        def decorator(f: Callable[..., Any]) -> Group:
            cmd = group(*args, **kwargs)(f)
            self.add_command(cmd)
            return cmd

        return decorator

    def list_commands(self, ctx: click.Context) -> Iterable[str]:
        return sorted(self.commands)


def group(name: Optional[str] = None, **kwargs: Any) -> Group:
    kwargs.setdefault("cls", Group)
    return click.group(name=name, **kwargs)  # type: ignore


@group()
def topics() -> None:
    """Help topics."""


@topics.command()
async def ps_format() -> None:
    """**Format for columns specification.**

      The format is a sequence of column specifications separated
      by commas or spaces: `{id}, {status}, {when}`.

      A column spec has a mandatory column id plus optional properties
      for indication of alignment, minimum and maximum column width,
      and optional column title:

        `{id;align=center;min=10;max=30;width=20;TITLE}`

      Here **id** is the column id, **align**, **min**, **max**, **width**
      are properties and **TITLE** is the column title.

      An alternative form is specifying the column id only without
      additional properties, in this case curly brackets can be omitted:
      `id, status, when` or `id status when` are valid formats.


    Available properties:

        **align**  Column aligning, accepted values: left, right, center.
        **min**    Minimal column width.
        **max**    Maximal column width.
        **width**  Default column width.

      All properties can be skipped, the default value for specified column ID
      is used in this case.


    The system recognizes the following columns:

      **ID**            **TITLE**       **ALIGN** **MIN** **MAX**  **WIDTH**
      ---------------------------------------------
      id            ID          left  -   -    -
      name          NAME        left  -   40   -
      tags          TAGS        left  -   40   -
      status        STATUS      left  -   10   -
      when          WHEN        left  -   15   -
      created       CREATED     left  -   15   -
      started       STARTED     left  -   15   -
      finished      FINISHED    left  -   15   -
      image         IMAGE       left  -   40   -
      owner         OWNER       left  -   25   -
      cluster_name  CLUSTER     left  -   15   -
      description   DESCRIPTION left  -   50   -
      command       COMMAND     left  -   100  -
      life_span     LIFE-SPAN   left  -   -    -

      By default all columns are left aligned and have no minimal and default widths.

      The column id is case insensitive, it can be shrinked to any unambiguous subset of
      the full name.  For example `{CLUSTER:max=20}` is a good column spec but
      `{C:max=20}` is not; it can be expanded into both `cluster_name` and `command`
      column ids.

    """


@topics.command()
async def user_config() -> None:
    """\
    **User configuration files.**

      The Neuro client supports user configuration files to provide default values
      for particular command options, user defined command aliases etc.

      There are two configuration files: **global** and **local**, both are optional
      and can be absent.

      The global file is located in the standard neuro config path.  "neuro" CLI uses
      `~/.neuro` folder by default, the path for global config file is
      `~/.neuro/user.toml`.

      The local config file is named .neuro.toml, the CLI search for this file
      starting from the current folder up to the root directory.

      Found local and global configurations are merged.
      If a parameter is present are both global and local versions the local parameter
      take a precedence.

      Configuration files have a TOML format (a stricter version of well-known INI
      format). See `https://en.wikipedia.org/wiki/TOML` and
      `https://github.com/toml-lang/toml#toml` for the format specification details.

    Supported configuration sections and parameters:

    **[alias]**

      A section for describing user-provided aliases.

      See `neuro help aliases` for details about avaiable section contents.

    **[job]**

      A section for `neuro job` command group settings.

    **ps-format**

      Default value for `neuro ps --format=XXX` option.

      See `neuro help ps-format` for information about the value specification.

    **life-span**

      Default job run-time limit for `neuro run --life-span=XXX` option.

      The value is a string in format `1d2h3m4s` (this example will set the limit to
      1 day, 2 hours, 3 minutes and 4 seconds). Some values can be missing, for example:
      `1d6h`, `30m`. No spaces are allowed between values.

      To completely disable run-time limit, use `0`.

    **[storage]**

      A section for `neuro storage` command group settings.

    **cp-exclude**

      Default value for `neuro cp --exclude=XXX` and `neuro cp --include=YYY` options.

      The value is a list of shell wildcard patterns, a file or folder that matches a
      pattern is excluded from processing.

      The pattern can contain `*` and `?`, e.g. `["*.jpg"]` is for exclusion of all
      files with `.jpg` extension.

      Exclamation mark ! is used to negate the pattern, e.g. `["*.jpg", "!main.jpg"]`
      excludes all `.jpg` files except `main.jpg`.

    **cp-exclude-from-files**

      Default value for `neuro cp --exclude-from-files=XXX` option.

      The value is a list of filenames that contain patterns for exclusion files
      and directories from uploading. For every proceeded folder
      patterns from matched exclusion files (e.g. ".neuroignore")
      are read and recursively applied to the directory content.
      Default is `[".neuroignore"]`.

      The format of files is the same as the format of `.gitignore` files:
      every line contains a pattern, exclamation mark `!` is used to negate
      the pattern, empty lines and lines which start with `#` are ignored.

    **[disk]**

      A section for `neuro disk` command group settings.

    **life-span**

      Default disk lifetime limit for `neuro disk create --life-span=XXX` option.

      The value is a string in format `1d2h3m4s` (this example will set the limit to
      1 day, 2 hours, 3 minutes and 4 seconds). Some values can be missing, for example:
      `1d6h`, `30m`. No spaces are allowed between values.

      To completely disable run-time limit, use `0`.

    Example:
    ```
      # jobs section
      [job]
      ps-format = "{id;max=30}, {status;max=10}"
      life-span = "1d6h"

      # storage section
      [storage]
      cp-exclude = ["*.jpg", "!main.jpg"]
      cp-exclude-from-files = [".neuroignore", ".gitignore"]

      # jobs section
      [disk]
      life-span = "7d"
    ```

    """


@topics.command()
async def aliases() -> None:
    """\
    **Custom command aliases.**

      Aliases exist to provide for abbreviating a system command,
      or for adding default arguments to a regularly used command.

      Aliases are described in user-config files
      (see `neuro help user-config` for details).

      `~/.neuro/user.toml` is used for **global** aliases,
      `.neuro.toml` can be used for saving **project-specific** aliases.
      Project aliases everrides global ones if the same alias
      name exists in both configuration files.

      There are **internal** and **external** aliases.
      An **internal** alias executes built-in neuro command in-place, an **
      external** alias executes any **system OS** command.

    **Internal alias**

      The internal alias is used for running existing neuro CLI command under
      a different name and with optional overriden defaults (passed predefined
      command-line options and arguments).

      For example, the following alias definition makes `neuro lsl` command
      that executes `neuro storage ls -hl` for listing the storage content
      using a long output mode with human-readable file sizes.

    ```
      [alias.lsl]
      cmd = "ls -l --human-readable"
      help = "List directory contents in a long mode.
    ```

      Available configuration arguments:

      * `[alias.lsl]` -- defines a subgroup for named alias,
                       `lsl` in this case.
      * `cmd`         -- command to execute with provided overridden options,
                       the key is **mandatory**.
                      `cmd` key in alias section implies **internal alias** mode.
      * `help`        -- help string, displayed by `neuro du --help`
                       command (optional),

      Internal allases accept additional command line options and agruments,
      these parameters are passed to underlying command as is.

      E.g., `neuro lsl storage:directory` works as
      `neuro ls -l --human-readable storage:directory`


    **External alias**

      The external alias spawns a subprocess with passing default options and
      arguments, all user-provided arguments are passed to underlying
      programm as well.

      For example, the following configuration defines `neuro du` command as
      an alias for system `du --human-readable` with optional providing the
      directory for analyzing.

    ```
      [alias.du]
      exec = "du"
      args = "[FILE]..."
      options = [
        "-h, --human-readable   print sizes in powers of 1024 (e.g., 1023M)",
        "-d, --max-depth=N  max recursion level for subdirectories lookup",
      ]
      help = '''
        Summarize disk usage of the set of FILEs,
        recursively for directories.
      '''
    ```

      Available configuration arguments:

      * `[alias.du]` -- defines a subgroup for named alias,
                      `du` in this case.
      * `exec`        -- external command to execute, the key is **mandatory**.
                      `exec` key in alias section implies **external alias** mode.
      * `args`        -- positional args accepted by the alias,
                       the format is described below (optional).
      * `options`     -- options and flags accepted by the alias,
                       the format is described below (optional).
      * `help`       -- help string, displayed by `neuro lsl --help`
                       command (optional),

      **args** is string with sequence of arguments, e.g. `DIR SRC... [DST]`

        If an argument is enclosed in brackets it is **optional** (`[FILE]`).
        If an argument is ended with ellipsis the argument accepts
        multiple values (`SRC...`)

      **options** is a list of strings, a string per option.

        Each string describes a single option, the options definition is separated
        from the option description (help) by two or more spaces.

        The option definition can contain
        * short name (`-h`)
        * long name (`--human-readable`)
        * indication for required value (`-d, --max-depth=N`)
          If the required value indicator (`=NAME`) is absent
          the option is considered as boolean flag.

      **exec** defines an external system to execute.

        The command is spawn in a subprocess, Neuro CLI waits for the subprocess
        finish, and, in turn, returns the exit code to the outer caller.

        The parameter may specify and executable file along with some options,
        e.g. `exec = "du --human-readable"` enforces human-readable mode
        for `du` command.

      `exec` can be used in **simplified** and **pattern** mode.

    **Pattern mode**

      In **pattern mode** the system command is used along with **substitutions**,
      e.g. `exec = "du {human_readable} {max_depth} {file}"`.
      Substitution is a variable name to expand enclosed in figure brackets,
      e.g. `{file}`.

      It is expanded with an option or positional argument specified
      by `args` or `options`.  The substitution name is automatically lowercased,
      minus (`-`) is replaced with underscore (`_`).
      E.g. `args = "ARG-NAME"` matches to `{arg_name}`.

      If the substitution corresponds to optional parameter and it is not provided
      by user the substitution is expanded to empty string.

      If the substitution corresponds to multiple values all of them are provided,
      e.g. `neuro du folder1 folder2` expands to `du folder1 folder2` since
      `[FILE]...` argument matches to `folder1 folder2` values.

      Options are expanded using longest form if provided,
      e.g. `neuro du -h` is expanded to `du --human-readable`.

      Options with values are expanded as well,
      e.g. `neuro du -d 1` is expanded to `du --max-depth 1`,
      `neuro du --max-depth 1` matches to the same command.

    **Simplified mode**

      In **simplified mode** the `exec` value does not contain any **substitutions**.
      In this case all parsed `options` and `args` are appended
      to executed command automatically if provided,
      e.g. `exec = "du"` is expanded to
      `exec = "du {human_readable} {max_depth} {file}"`

    """


@topics.command()
async def secrets() -> None:
    """\
    **Using secrets.**

      Secret is a named encrypted data stored in the Neuro Platform Cluster.

      A user can create a secret, list available secret names and delete unused secrets
      but the reading of secret's data back is forbidden.  Instead of bare reading,
      secrets can be accessed from a running job as environment variable or mounted
      file.

      Secrets are isolated and user-specific, a secret that belongs to user A cannot be
      accessed by user B.

    **Secrets management**

      Use `neuro secret` command group for managing secrets.

      `neuro secret ls` prints all available secret names.

      `neuro secret add key value` creates a secret named *key* with encrypted data
      *value*.

      To store the file's content as a secret please use
      `neuro secret add KEY_NAME @path/to/file.txt` notation.

      `neuro secret rm key` removes the secret *key*.

      Internally, Neuro Platform uses Kubernetes Cluster secrets subsystem a secrets
      storage.

    **Secrets usage**

      As said above, you cannot read a secret directly but should pass it into a running
      job as an environment variable or mounted file.

      To pass a secret *key* as environment variable please use `secret:key` as a value,
      e.g. `neuro run --env VAR=secret:key ...` form.

      To mount a secret as a file please use `secret:` volume's schema, e.g.
      `neuro run --volume secret:key:/mount/path/file.txt`.

    """



import contextlib
import logging
import re
from dataclasses import replace
from typing import Any, Dict, List, Optional, Set

import aiodocker
import aiohttp
from aiodocker.exceptions import DockerError

from .abc import (
    AbstractDockerImageProgress,
    ImageCommitFinished,
    ImageCommitStarted,
    ImageProgressPull,
    ImageProgressPush,
    ImageProgressSave,
    ImageProgressStep,
)
from .config import Config
from .core import _Core
from .errors import AuthorizationError
from .parser import Parser
from .parsing_utils import LocalImage, RemoteImage, TagOption, _as_repo_str
from .utils import NoPublicConstructor


REPOS_PER_PAGE = 30
TAGS_PER_PAGE = 30

log = logging.getLogger(__name__)


class Images(metaclass=NoPublicConstructor):
    def __init__(self, core: _Core, config: Config, parse: Parser) -> None:
        self._core = core
        self._config = config
        self._parse = parse
        self._temporary_images: Set[str] = set()
        self.__docker: Optional[aiodocker.Docker] = None
        self._registry_url = self._config.registry_url.with_path("/v2/")

    @property
    def _docker(self) -> aiodocker.Docker:
        if not self.__docker:
            try:
                self.__docker = aiodocker.Docker()
            except ValueError as error:
                if re.match(
                    r".*Either DOCKER_HOST or local sockets are not available.*",
                    f"{error}",
                ):
                    raise DockerError(
                        900,
                        {
                            "message": "Docker engine is not available. "
                            "Please specify DOCKER_HOST variable "
                            "if you are using remote docker engine"
                        },
                    )
                raise
        return self.__docker

    async def _close(self) -> None:
        for image in self._temporary_images:
            with contextlib.suppress(DockerError, aiohttp.ClientError):
                await self._docker.images.delete(image)
        if self.__docker is not None:
            await self.__docker.close()

    async def push(
        self,
        local: LocalImage,
        remote: Optional[RemoteImage] = None,
        *,
        progress: Optional[AbstractDockerImageProgress] = None,
    ) -> RemoteImage:
        if remote is None:
            remote = self._parse._local_to_remote_image(local)

        if progress is None:
            progress = _DummyProgress()
        progress.push(ImageProgressPush(local, remote))

        repo = _as_repo_str(remote)
        try:
            await self._docker.images.tag(str(local), repo)
        except DockerError as error:
            if error.status == 404:
                raise ValueError(
                    f"Image {local} was not found " "in your local docker images"
                ) from error
        auth = await self._config._docker_auth()
        try:
            async for obj in self._docker.images.push(repo, auth=auth, stream=True):
                step = _try_parse_image_progress_step(obj, remote.tag)
                if step:
                    progress.step(step)
        except DockerError as error:
            # TODO check this part when registry fixed
            if error.status == 403:
                raise AuthorizationError(f"Access denied {remote}") from error
            raise  # pragma: no cover
        return remote

    async def pull(
        self,
        remote: RemoteImage,
        local: Optional[LocalImage] = None,
        *,
        progress: Optional[AbstractDockerImageProgress] = None,
    ) -> LocalImage:
        if local is None:
            local = self._parse._remote_to_local_image(remote)

        if progress is None:
            progress = _DummyProgress()
        progress.pull(ImageProgressPull(remote, local))

        repo = _as_repo_str(remote)
        auth = await self._config._docker_auth()
        try:
            async for obj in self._docker.pull(repo, auth=auth, repo=repo, stream=True):
                self._temporary_images.add(repo)
                step = _try_parse_image_progress_step(obj, remote.tag)
                if step:
                    progress.step(step)
        except DockerError as error:
            if error.status == 404:
                raise ValueError(
                    f"Image {remote} was not found " "in registry"
                ) from error
            # TODO check this part when registry fixed
            elif error.status == 403:
                raise AuthorizationError(f"Access denied {remote}") from error
            raise  # pragma: no cover

        await self._docker.images.tag(repo, str(local))

        return local

    async def ls(self) -> List[RemoteImage]:
        auth = await self._config._registry_auth()
        prefix = f"image://{self._config.cluster_name}/"
        url = self._registry_url / "_catalog"
        result: List[RemoteImage] = []
        while True:
            url = url.update_query(n=str(REPOS_PER_PAGE))
            async with self._core.request("GET", url, auth=auth) as resp:
                ret = await resp.json()
                repos = ret["repositories"]
                for repo in repos:
                    try:
                        result.append(
                            self._parse.remote_image(
                                prefix + repo, tag_option=TagOption.DENY
                            )
                        )
                    except ValueError as err:
                        log.warning(str(err))
                if not repos or "next" not in resp.links:
                    break
                url = resp.links["next"]["url"]
        return result

    def _validate_image_for_tags(self, image: RemoteImage) -> None:
        err = f"Invalid image `{image}`: "
        if image.tag is not None:
            raise ValueError(err + "tag is not allowed")
        if not image.owner:
            raise ValueError(err + "missing image owner")
        if not image.name:
            raise ValueError(err + "missing image name")

    async def tags(self, image: RemoteImage) -> List[RemoteImage]:
        self._validate_image_for_tags(image)
        name = f"{image.owner}/{image.name}"
        auth = await self._config._registry_auth()
        url = self._registry_url / name / "tags" / "list"
        result: List[RemoteImage] = []
        while True:
            url = url.update_query(n=str(TAGS_PER_PAGE))
            async with self._core.request("GET", url, auth=auth) as resp:
                ret = await resp.json()
                tags = ret.get("tags", [])
                for tag in tags:
                    result.append(replace(image, tag=tag))
                if not tags or "next" not in resp.links:
                    break
                url = resp.links["next"]["url"]
        return result


def _try_parse_image_progress_step(
    obj: Dict[str, Any], target_image_tag: Optional[str]
) -> Optional[ImageProgressStep]:
    _raise_on_error_chunk(obj)
    if "id" in obj.keys() and obj["id"] != target_image_tag:
        if "progress" in obj.keys():
            message = f"{obj['id']}: {obj['status']} {obj['progress']}"
        else:
            message = f"{obj['id']}: {obj['status']}"
        return ImageProgressStep(message, obj["id"])
    return None


def _raise_on_error_chunk(obj: Dict[str, Any]) -> None:
    if "error" in obj.keys():
        error_details = obj.get("errorDetail", {"message": "Unknown error"})
        raise DockerError(900, error_details)


class _DummyProgress(AbstractDockerImageProgress):
    def pull(self, data: ImageProgressPull) -> None:
        pass

    def push(self, data: ImageProgressPush) -> None:
        pass

    def step(self, data: ImageProgressStep) -> None:
        pass

    def save(self, data: ImageProgressSave) -> None:
        pass

    def commit_started(self, data: ImageCommitStarted) -> None:
        pass

    def commit_finished(self, data: ImageCommitFinished) -> None:
        pass



import sys
from ctypes import c_ulong


# Input/Output standard device numbers. Note that these are not handle objects.
# It's the `windll.kernel32.GetStdHandle` system call that turns them into a
# real handle object.
STD_INPUT_HANDLE = c_ulong(-10)
STD_OUTPUT_HANDLE = c_ulong(-11)
STD_ERROR_HANDLE = c_ulong(-12)

# See: https://msdn.microsoft.com/pl-pl/library/windows/desktop/ms686033(v=vs.85).aspx
ENABLE_PROCESSED_INPUT = 0x0001
ENABLE_VIRTUAL_TERMINAL_PROCESSING = 0x0004


def enable_ansi() -> None:
    if not sys.stdout.isatty():
        return
    if sys.platform == "win32":
        from ctypes import windll
        from ctypes.wintypes import DWORD, HANDLE

        hconsole = HANDLE(windll.kernel32.GetStdHandle(STD_OUTPUT_HANDLE))

        # Enable processing of vt100 sequences.
        result = windll.kernel32.SetConsoleMode(
            hconsole,
            DWORD(ENABLE_PROCESSED_INPUT | ENABLE_VIRTUAL_TERMINAL_PROCESSING),
        )

        if result != 1:
            raise RuntimeError("Windows 10 is required.")
    else:
        return



import argparse
import sys


print(sys.argv[1:])


parser = argparse.ArgumentParser()
parser.add_argument("--exit", type=int, default=0)


args, unparsed = parser.parse_known_args()
sys.exit(args.exit)



import os
import sys
from pathlib import Path

import click

from .root import Root
from .utils import argument, group


CFG_FILE = {"bash": Path("~/.bashrc"), "zsh": Path("~/.zshrc")}
SOURCE_CMD = {"bash": "source", "zsh": "source_zsh"}

ACTIVATION_TEMPLATE = 'eval "$(_NEURO_COMPLETE={cmd} {exe})"'


@group()
def completion() -> None:
    """
    Output shell completion code.
    """


@completion.command()
@argument("shell", type=click.Choice(["bash", "zsh"]))
async def generate(root: Root, shell: str) -> None:
    """
    Provide an instruction for shell completion generation.
    """
    click.echo(f"Push the following line into your {CFG_FILE[shell]}")
    click.echo(ACTIVATION_TEMPLATE.format(cmd=SOURCE_CMD[shell], exe=sys.argv[0]))


@completion.command()
@argument("shell", type=click.Choice(["bash", "zsh"]))
async def patch(root: Root, shell: str) -> None:
    """
    Automatically patch shell configuration profile to enable completion
    """
    profile_file = CFG_FILE[shell].expanduser()
    with profile_file.open("ab+") as profile:
        profile.write(
            b"\n"
            + os.fsencode(
                ACTIVATION_TEMPLATE.format(cmd=SOURCE_CMD[shell], exe=sys.argv[0])
            )
            + b"\n"
        )



import operator
from typing import Iterable, Iterator, List

import click
from click import style

from neuromation.api.admin import _Cluster, _ClusterUser, _NodePool
from neuromation.cli.utils import format_size

from .ftable import Align, table


class ClusterUserFormatter:
    def __call__(self, clusters_users: Iterable[_ClusterUser]) -> List[str]:
        headers = (click.style("Name", bold=True), click.style("Role", bold=True))
        rows = []

        for user in clusters_users:
            rows.append((user.user_name, user.role.value))
        rows.sort(key=operator.itemgetter(0))

        rows.insert(0, headers)
        return list(table(rows=rows))


class ClustersFormatter:
    def __call__(self, clusters: Iterable[_Cluster]) -> List[str]:
        out = []
        for cluster in clusters:
            prefix = "  "
            out.append(style(f"{cluster.name}:", bold=True))
            out.append(
                prefix + style("Status: ", bold=True) + cluster.status.capitalize()
            )
            if cluster.cloud_provider:
                cloud_provider = cluster.cloud_provider
                if cloud_provider.type != "on_prem":
                    out.append(
                        prefix + style("Cloud: ", bold=True) + cloud_provider.type
                    )
                if cloud_provider.region:
                    out.append(
                        prefix + style("Region: ", bold=True) + cloud_provider.region
                    )
                if cloud_provider.zones:
                    out.append(
                        prefix
                        + style("Zones: ", bold=True)
                        + ", ".join(cloud_provider.zones)
                    )
                if cloud_provider.node_pools:
                    out.append(prefix + style("Node pools:", bold=True))
                    out.extend(
                        _format_node_pools(cloud_provider.node_pools, prefix + "  ")
                    )
                if cloud_provider.storage:
                    out.append(
                        prefix
                        + style("Storage: ", bold=True)
                        + cloud_provider.storage.description
                    )
        return out


def _format_node_pools(node_pools: Iterable[_NodePool], prefix: str) -> Iterator[str]:
    is_scalable = _is_scalable(node_pools)
    has_preemptible = _has_preemptible(node_pools)
    has_tpu = _has_tpu(node_pools)
    has_idle = _has_idle(node_pools)

    headers = ["Machine", "CPU", "Memory"]
    if has_preemptible:
        headers.append("Preemptible")
    headers.append("GPU")
    if has_tpu:
        headers.append("TPU")
    if is_scalable:
        headers.append("Min")
        headers.append("Max")
    else:
        headers.append("Size")
    if has_idle:
        headers.append("Idle")

    rows = [headers]
    for node_pool in node_pools:
        row = [
            node_pool.machine_type,
            str(node_pool.available_cpu),
            format_size(node_pool.available_memory_mb * 1024 ** 2),
        ]
        if has_preemptible:
            row.append("√" if node_pool.is_preemptible else "×")
        row.append(_gpu(node_pool))
        if has_tpu:
            row.append("√" if node_pool.is_tpu_enabled else "×")
        if is_scalable:
            row.append(str(node_pool.min_size))
        row.append(str(node_pool.max_size))
        if has_idle:
            row.append(str(node_pool.idle_size))
        rows.append(row)

    aligns = [Align.LEFT, Align.RIGHT, Align.RIGHT]
    if has_preemptible:
        aligns.append(Align.CENTER)
    aligns.append(Align.RIGHT)
    if has_tpu:
        aligns.append(Align.CENTER)
    aligns.append(Align.RIGHT)
    if is_scalable:
        aligns.append(Align.RIGHT)
    if has_idle:
        aligns.append(Align.RIGHT)

    for line in table(rows=rows, aligns=aligns):
        yield prefix + line


def _is_scalable(node_pools: Iterable[_NodePool]) -> bool:
    for node_pool in node_pools:
        if node_pool.min_size != node_pool.max_size:
            return True
    return False


def _has_preemptible(node_pools: Iterable[_NodePool]) -> bool:
    for node_pool in node_pools:
        if node_pool.is_preemptible:
            return True
    return False


def _has_tpu(node_pools: Iterable[_NodePool]) -> bool:
    for node_pool in node_pools:
        if node_pool.is_tpu_enabled:
            return True
    return False


def _has_idle(node_pools: Iterable[_NodePool]) -> bool:
    for node_pool in node_pools:
        if node_pool.idle_size:
            return True
    return False


def _gpu(node_pool: _NodePool) -> str:
    if node_pool.gpu:
        return f"{node_pool.gpu} x {node_pool.gpu_model}"
    return ""



import asyncio
import os
import ssl
import sys
from pathlib import Path
from typing import Awaitable, Callable, List, Optional

import aiohttp
import certifi
from yarl import URL

import neuromation

from .client import Client
from .config import _ConfigData, _load, _save
from .core import DEFAULT_TIMEOUT
from .errors import ConfigError
from .login import AuthNegotiator, HeadlessNegotiator, _AuthToken, logout_from_browser
from .server_cfg import _ServerConfig, get_server_config
from .tracing import _make_trace_config
from .utils import _ContextManager


DEFAULT_CONFIG_PATH = "~/.neuro"
CONFIG_ENV_NAME = "NEUROMATION_CONFIG"
DEFAULT_API_URL = URL("https://staging.neu.ro/api/v1")


def _make_session(
    timeout: aiohttp.ClientTimeout, trace_configs: Optional[List[aiohttp.TraceConfig]]
) -> _ContextManager[aiohttp.ClientSession]:
    return _ContextManager[aiohttp.ClientSession](
        __make_session(timeout, trace_configs)
    )


async def __make_session(
    timeout: aiohttp.ClientTimeout, trace_configs: Optional[List[aiohttp.TraceConfig]]
) -> aiohttp.ClientSession:
    ssl_context = ssl.SSLContext()
    ssl_context.load_verify_locations(capath=certifi.where())
    connector = aiohttp.TCPConnector(ssl=ssl_context)
    return aiohttp.ClientSession(
        timeout=timeout,
        connector=connector,
        trace_configs=trace_configs,
        headers={"User-Agent": f"NeuroCLI/{neuromation.__version__} ({sys.platform})"},
    )


class Factory:
    def __init__(
        self,
        path: Optional[Path] = None,
        trace_configs: Optional[List[aiohttp.TraceConfig]] = None,
        trace_id: Optional[str] = None,
    ) -> None:
        if path is None:
            path = Path(os.environ.get(CONFIG_ENV_NAME, DEFAULT_CONFIG_PATH))
        self._path = path.expanduser()
        self._trace_configs = [_make_trace_config()]
        if trace_configs:
            self._trace_configs += trace_configs
        self._trace_id = trace_id

    @property
    def path(self) -> Path:
        return self._path

    async def get(self, *, timeout: aiohttp.ClientTimeout = DEFAULT_TIMEOUT) -> Client:
        session = await _make_session(timeout, self._trace_configs)
        try:
            client = Client._create(session, self._path, self._trace_id)
            await client.config.check_server()
        except (asyncio.CancelledError, Exception):
            await session.close()
            raise
        else:
            return client

    async def login(
        self,
        show_browser_cb: Callable[[URL], Awaitable[None]],
        *,
        url: URL = DEFAULT_API_URL,
        timeout: aiohttp.ClientTimeout = DEFAULT_TIMEOUT,
    ) -> None:
        config_file = self._path / "db"
        if config_file.exists():
            raise ConfigError(f"Config at {self._path} already exists. Please logout")
        async with _make_session(timeout, self._trace_configs) as session:
            config_unauthorized = await get_server_config(session, url)
            negotiator = AuthNegotiator(
                session, config_unauthorized.auth_config, show_browser_cb
            )
            auth_token = await negotiator.get_token()

            config_authorized = await get_server_config(
                session, url, token=auth_token.token
            )
        config = self._gen_config(config_authorized, auth_token, url)
        self._save(config)

    async def login_headless(
        self,
        get_auth_code_cb: Callable[[URL], Awaitable[str]],
        *,
        url: URL = DEFAULT_API_URL,
        timeout: aiohttp.ClientTimeout = DEFAULT_TIMEOUT,
    ) -> None:
        config_file = self._path / "db"
        if config_file.exists():
            raise ConfigError(f"Config at {self._path} already exists. Please logout")
        async with _make_session(timeout, self._trace_configs) as session:
            config_unauthorized = await get_server_config(session, url)
            negotiator = HeadlessNegotiator(
                session, config_unauthorized.auth_config, get_auth_code_cb
            )
            auth_token = await negotiator.get_token()

            config_authorized = await get_server_config(
                session, url, token=auth_token.token
            )
        config = self._gen_config(config_authorized, auth_token, url)
        self._save(config)

    async def login_with_token(
        self,
        token: str,
        *,
        url: URL = DEFAULT_API_URL,
        timeout: aiohttp.ClientTimeout = DEFAULT_TIMEOUT,
    ) -> None:
        config_file = self._path / "db"
        if config_file.exists():
            raise ConfigError(f"Config at {self._path} already exists. Please logout")
        async with _make_session(timeout, self._trace_configs) as session:
            server_config = await get_server_config(session, url, token=token)
        config = self._gen_config(
            server_config, _AuthToken.create_non_expiring(token), url
        )
        self._save(config)

    def _gen_config(
        self, server_config: _ServerConfig, token: _AuthToken, url: URL
    ) -> _ConfigData:
        cluster_name = next(iter(server_config.clusters))
        config = _ConfigData(
            auth_config=server_config.auth_config,
            auth_token=token,
            url=url,
            version=neuromation.__version__,
            cluster_name=cluster_name,
            clusters=server_config.clusters,
        )
        return config

    async def logout(
        self,
        show_browser_cb: Callable[[URL], Awaitable[None]] = None,
    ) -> None:
        if show_browser_cb is not None:
            try:
                old_config = _load(self._path)
            except ConfigError:
                pass  # Do not try to logout from auth0 if config is broken
            else:
                await logout_from_browser(old_config.auth_config, show_browser_cb)

        files = ["db", "db-wal", "db-shm"]
        for name in files:
            f = self._path / name
            if f.exists():
                f.unlink()
        if self._path.is_file():
            # Old-styled single file config from 2019
            self._path.unlink()
        else:
            try:
                self._path.rmdir()
            except OSError:
                # Directory Not Empty or Not A Directory
                pass

    def _save(self, config: _ConfigData) -> None:
        _save(config, self._path)



import shutil
from pathlib import Path
from typing import Any, Callable, List

import asynctest

from neuromation.cli.utils import NEURO_STEAL_CONFIG

from .conftest import SysCapWithCode


_RunCli = Callable[[List[str]], SysCapWithCode]


@asynctest.mock.patch("neuromation.api.jobs.Jobs.get_capacity")
def test_steal_config(
    patched_jobs_counts: Any,
    nmrc_path: Path,
    run_cli: _RunCli,
    tmp_path: Path,
    monkeypatch: Any,
) -> None:
    patched_jobs_counts.return_value = {}
    folder = shutil.move(nmrc_path, tmp_path / "orig-cfg")  # type: ignore
    monkeypatch.setenv(NEURO_STEAL_CONFIG, str(folder))
    ret = run_cli(["--neuromation-config", str(tmp_path / ".neuro"), "config", "show"])
    assert ret.code == 0, ret


@asynctest.mock.patch("neuromation.api.jobs.Jobs.get_capacity")
def test_steal_config_dont_override_existing(
    patched_jobs_counts: Any,
    nmrc_path: Path,
    run_cli: _RunCli,
    tmp_path: Path,
    monkeypatch: Any,
) -> None:
    patched_jobs_counts.return_value = {}
    folder = shutil.copytree(nmrc_path, tmp_path / "orig-cfg")
    monkeypatch.setenv(NEURO_STEAL_CONFIG, str(folder))
    ret = run_cli(["--neuromation-config", str(nmrc_path), "config", "show"])
    # FileExistsError
    assert ret.code == 74, ret
    assert "FileExistsError" in ret.err



import asyncio
import dataclasses
import glob as globmodule  # avoid conflict with subcommand "glob"
import logging
import sys
from typing import Any, Callable, List, Optional, Sequence, Tuple

import click
from yarl import URL

from neuromation.api import (
    Client,
    FileStatusType,
    IllegalArgumentError,
    ResourceNotFound,
)
from neuromation.api.file_filter import FileFilter
from neuromation.api.url_utils import _extract_path

from .const import EX_OSFILE
from .formatters.storage import (
    BaseFilesFormatter,
    DeleteProgress,
    FilesSorter,
    LongFilesFormatter,
    SimpleFilesFormatter,
    Tree,
    TreeFormatter,
    VerticalColumnsFilesFormatter,
    create_storage_progress,
    get_painter,
)
from .root import Root
from .utils import (
    Option,
    argument,
    command,
    group,
    option,
    pager_maybe,
    parse_file_resource,
)


NEUROIGNORE_FILENAME = ".neuroignore"

log = logging.getLogger(__name__)


@group()
def storage() -> None:
    """
    Storage operations.
    """


@command()
@argument("paths", nargs=-1, required=True)
@option(
    "--recursive",
    "-r",
    is_flag=True,
    help="remove directories and their contents recursively",
)
@option(
    "--glob/--no-glob",
    is_flag=True,
    default=True,
    show_default=True,
    help="Expand glob patterns in PATHS",
)
@option(
    "-p/-P",
    "--progress/--no-progress",
    is_flag=True,
    default=None,
    help="Show progress, on by default in TTY mode, off otherwise.",
)
async def rm(
    root: Root,
    paths: Sequence[str],
    recursive: bool,
    glob: bool,
    progress: Optional[bool],
) -> None:
    """
    Remove files or directories.

    Examples:

    neuro rm storage:foo/bar
    neuro rm storage://{username}/foo/bar
    neuro rm --recursive storage://{username}/foo/
    neuro rm storage:foo/**/*.tmp
    """
    errors = False
    show_progress = root.tty if progress is None else progress

    for uri in await _expand(paths, root, glob):
        try:
            progress_obj = DeleteProgress(root) if show_progress else None
            await root.client.storage.rm(
                uri, recursive=recursive, progress=progress_obj
            )
        except (OSError, ResourceNotFound, IllegalArgumentError) as error:
            log.error(f"cannot remove {uri}: {error}")
            errors = True
        else:
            if root.verbosity > 0:
                painter = get_painter(root.color, quote=True)
                curi = painter.paint(str(uri), FileStatusType.FILE)
                click.echo(f"removed {curi}")
    if errors:
        sys.exit(EX_OSFILE)


@command()
@argument("paths", nargs=-1)
@option(
    "-a",
    "--all",
    "show_all",
    is_flag=True,
    help="do not ignore entries starting with .",
)
@option(
    "-d",
    "--directory",
    is_flag=True,
    help="list directories themselves, not their contents.",
)
@option(
    "--human-readable",
    "-h",
    is_flag=True,
    help="with -l print human readable sizes (e.g., 2K, 540M).",
)
@option("-l", "format_long", is_flag=True, help="use a long listing format.")
@option(
    "--sort",
    type=click.Choice(["name", "size", "time"]),
    default="name",
    help="sort by given field, default is name.",
)
async def ls(
    root: Root,
    paths: Sequence[str],
    human_readable: bool,
    format_long: bool,
    sort: str,
    directory: bool,
    show_all: bool,
) -> None:
    """
    List directory contents.

    By default PATH is equal user's home dir (storage:)
    """
    if not paths:
        paths = ["storage:"]
    uris = [parse_file_resource(path, root) for path in paths]

    errors = False
    for uri in uris:
        try:
            if directory:
                files = [await root.client.storage.stat(uri)]
            else:
                if root.verbosity > 0:
                    painter = get_painter(root.color, quote=True)
                    curi = painter.paint(str(uri), FileStatusType.DIRECTORY)
                    click.echo(f"List of {curi}:")

                files = [file async for file in root.client.storage.ls(uri)]
                files = sorted(files, key=FilesSorter(sort).key())
        except (OSError, ResourceNotFound, IllegalArgumentError) as error:
            log.error(f"cannot access {uri}: {error}")
            errors = True
        else:
            if format_long:
                formatter: BaseFilesFormatter = LongFilesFormatter(
                    human_readable=human_readable, color=root.color
                )
            else:
                if root.tty:
                    formatter = VerticalColumnsFilesFormatter(
                        width=root.terminal_size[0], color=root.color
                    )
                else:
                    formatter = SimpleFilesFormatter(root.color)

            if not show_all:
                files = [item for item in files if not item.name.startswith(".")]
            pager_maybe(formatter(files), root.tty, root.terminal_size)

    if errors:
        sys.exit(EX_OSFILE)


@command()
@argument("patterns", nargs=-1, required=False)
async def glob(root: Root, patterns: Sequence[str]) -> None:
    """
    List resources that match PATTERNS.
    """
    for pattern in patterns:
        uri = parse_file_resource(pattern, root)
        if root.verbosity > 0:
            painter = get_painter(root.color, quote=True)
            curi = painter.paint(str(uri), FileStatusType.FILE)
            click.echo(f"Using pattern {curi}:")
        async for file in root.client.storage.glob(uri):
            click.echo(file)


class FileFilterParserOption(click.parser.Option):
    def process(self, value: str, state: click.parser.ParsingState) -> None:
        super().process((self.const, value), state)


class FileFilterOption(Option):
    def add_to_parser(self, parser: click.parser.OptionParser, ctx: Any) -> None:
        option = FileFilterParserOption(
            self.opts,
            self.name,
            action="append",
            nargs=self.nargs,
            const=self.flag_value,
            obj=self,
        )
        parser._opt_prefixes.update(option.prefixes)
        for opt in option._short_opts:
            parser._short_opt[opt] = option
        for opt in option._long_opts:
            parser._long_opt[opt] = option


def filter_option(*args: str, flag_value: bool, help: str) -> Callable[[Any], Any]:
    return option(
        *args,
        multiple=True,
        cls=FileFilterOption,
        flag_value=flag_value,
        type=click.UNPROCESSED,
        help=help,
        secure=True,
    )


@command()
@argument("sources", nargs=-1, required=False)
@argument("destination", required=False)
@option("-r", "--recursive", is_flag=True, help="Recursive copy, off by default")
@option(
    "--glob/--no-glob",
    is_flag=True,
    default=True,
    show_default=True,
    help="Expand glob patterns in SOURCES with explicit scheme.",
)
@option(
    "-t",
    "--target-directory",
    metavar="DIRECTORY",
    default=None,
    help="Copy all SOURCES into DIRECTORY.",
)
@option(
    "-T",
    "--no-target-directory",
    is_flag=True,
    help="Treat DESTINATION as a normal file.",
)
@option(
    "-u",
    "--update",
    is_flag=True,
    help="Copy only when the SOURCE file is newer than the destination file "
    "or when the destination file is missing.",
)
@filter_option(
    "--exclude",
    "filters",
    flag_value=True,
    help=("Exclude files and directories that match the specified pattern."),
)
@filter_option(
    "--include",
    "filters",
    flag_value=False,
    help=("Don't exclude files and directories that match the specified pattern."),
)
@option(
    "--exclude-from-files",
    metavar="FILES",
    help=(
        "A list of file names that contain patterns for exclusion files "
        "and directories. Used only for uploading. "
        "The default can be changed using the storage.cp-exclude-from-files "
        'configuration variable documented in "neuro help user-config"'
    ),
)
@option(
    "-p/-P",
    "--progress/--no-progress",
    is_flag=True,
    default=True,
    help="Show progress, on by default in TTY mode, off otherwise.",
)
async def cp(
    root: Root,
    sources: Sequence[str],
    destination: Optional[str],
    recursive: bool,
    glob: bool,
    target_directory: Optional[str],
    no_target_directory: bool,
    update: bool,
    filters: Optional[Tuple[Tuple[bool, str], ...]],
    exclude_from_files: str,
    progress: bool,
) -> None:
    """
    Copy files and directories.

    Either SOURCES or DESTINATION should have storage:// scheme.
    If scheme is omitted, file:// scheme is assumed.

    Use /dev/stdin and /dev/stdout file names to copy a file from terminal
    and print the content of file on the storage to console.

    Any number of --exclude and --include options can be passed.  The
    filters that appear later in the command take precedence over filters
    that appear earlier in the command.  If neither --exclude nor
    --include options are specified the default can be changed using the
    storage.cp-exclude configuration variable documented in
    "neuro help user-config".

    Examples:

    # copy local files into remote storage root
    neuro cp foo.txt bar/baz.dat storage:
    neuro cp foo.txt bar/baz.dat -t storage:

    # copy local directory `foo` into existing remote directory `bar`
    neuro cp -r foo -t storage:bar

    # copy the content of local directory `foo` into existing remote
    # directory `bar`
    neuro cp -r -T storage:foo storage:bar

    # download remote file `foo.txt` into local file `/tmp/foo.txt` with
    # explicit file:// scheme set
    neuro cp storage:foo.txt file:///tmp/foo.txt
    neuro cp -T storage:foo.txt file:///tmp/foo.txt
    neuro cp storage:foo.txt file:///tmp
    neuro cp storage:foo.txt -t file:///tmp

    # download other user's remote file into the current directory
    neuro cp storage://{username}/foo.txt .

    # download only files with extension `.out` into the current directory
    neuro cp storage:results/*.out .
    """
    target_dir: Optional[URL]
    dst: Optional[URL]
    if target_directory:
        if no_target_directory:
            raise click.UsageError(
                "Cannot combine --target-directory (-t) and --no-target-directory (-T)"
            )
        if destination is None:
            raise click.MissingParameter(
                param_type="argument", param_hint='"SOURCES..."'
            )
        sources = *sources, destination
        target_dir = parse_file_resource(target_directory, root)
        dst = None
    else:
        if destination is None:
            raise click.MissingParameter(
                param_type="argument", param_hint='"DESTINATION"'
            )
        if not sources:
            raise click.MissingParameter(
                param_type="argument", param_hint='"SOURCES..."'
            )
        dst = parse_file_resource(destination, root)
        if no_target_directory or not await _is_dir(root, dst):
            target_dir = None
        else:
            target_dir = dst
            dst = None

    ignore_file_names = await calc_ignore_file_names(root.client, exclude_from_files)
    filters = await calc_filters(root.client, filters)
    srcs = await _expand(sources, root, glob, allow_file=True)
    if no_target_directory and len(srcs) > 1:
        raise click.UsageError(f"Extra operand after {str(srcs[1])!r}")

    file_filter = FileFilter()
    for exclude, pattern in filters:
        log.debug("%s %s", "Exclude" if exclude else "Include", pattern)
        file_filter.append(exclude, pattern)

    show_progress = root.tty and progress

    errors = False
    for src in srcs:
        if target_dir:
            dst = target_dir / src.name
        assert dst

        progress_obj = create_storage_progress(root, show_progress)
        progress_obj.begin(src, dst)

        try:
            if src.scheme == "file" and dst.scheme == "storage":
                if recursive and await _is_dir(root, src):
                    await root.client.storage.upload_dir(
                        src,
                        dst,
                        update=update,
                        filter=file_filter.match,
                        ignore_file_names=frozenset(ignore_file_names),
                        progress=progress_obj,
                    )
                else:
                    await root.client.storage.upload_file(
                        src, dst, update=update, progress=progress_obj
                    )
            elif src.scheme == "storage" and dst.scheme == "file":
                if recursive and await _is_dir(root, src):
                    await root.client.storage.download_dir(
                        src,
                        dst,
                        update=update,
                        filter=file_filter.match,
                        progress=progress_obj,
                    )
                else:
                    await root.client.storage.download_file(
                        src, dst, update=update, progress=progress_obj
                    )
            else:
                raise RuntimeError(
                    f"Copy operation of the file with scheme '{src.scheme}'"
                    f" to the file with scheme '{dst.scheme}'"
                    f" is not supported"
                )
        except (OSError, ResourceNotFound, IllegalArgumentError) as error:
            log.error(f"cannot copy {src} to {dst}: {error}")
            errors = True

        progress_obj.end()

    if errors:
        sys.exit(EX_OSFILE)


@command()
@argument("paths", nargs=-1, required=True)
@option(
    "-p",
    "--parents",
    is_flag=True,
    help="No error if existing, make parent directories as needed",
)
async def mkdir(root: Root, paths: Sequence[str], parents: bool) -> None:
    """
    Make directories.
    """
    uris = [parse_file_resource(path, root) for path in paths]

    errors = False
    for uri in uris:
        try:
            await root.client.storage.mkdir(uri, parents=parents, exist_ok=parents)
        except (OSError, ResourceNotFound, IllegalArgumentError) as error:
            log.error(f"cannot create directory {uri}: {error}")
            errors = True
        else:
            if root.verbosity > 0:
                painter = get_painter(root.color, quote=True)
                curi = painter.paint(str(uri), FileStatusType.DIRECTORY)
                click.echo(f"created directory {curi}")

    if errors:
        sys.exit(EX_OSFILE)


@command()
@argument("sources", nargs=-1, required=False)
@argument("destination", required=False)
@option(
    "--glob/--no-glob",
    is_flag=True,
    default=True,
    show_default=True,
    help="Expand glob patterns in SOURCES",
)
@option(
    "-t",
    "--target-directory",
    metavar="DIRECTORY",
    default=None,
    help="Copy all SOURCES into DIRECTORY",
)
@option(
    "-T",
    "--no-target-directory",
    is_flag=True,
    help="Treat DESTINATION as a normal file",
)
async def mv(
    root: Root,
    sources: Sequence[str],
    destination: Optional[str],
    glob: bool,
    target_directory: Optional[str],
    no_target_directory: bool,
) -> None:
    """
    Move or rename files and directories.

    SOURCE must contain path to the
    file or directory existing on the storage, and DESTINATION must contain
    the full path to the target file or directory.

    Examples:

    # move and rename remote file
    neuro mv storage:foo.txt storage:bar/baz.dat
    neuro mv -T storage:foo.txt storage:bar/baz.dat

    # move remote files into existing remote directory
    neuro mv storage:foo.txt storage:bar/baz.dat storage:dst
    neuro mv storage:foo.txt storage:bar/baz.dat -t storage:dst

    # move the content of remote directory into other existing
    # remote directory
    neuro mv -T storage:foo storage:bar

    # move remote file into other user's directory
    neuro mv storage:foo.txt storage://{username}/bar.dat

    # move remote file from other user's directory
    neuro mv storage://{username}/foo.txt storage:bar.dat
    """
    target_dir: Optional[URL]
    dst: Optional[URL]
    if target_directory:
        if no_target_directory:
            raise click.UsageError(
                "Cannot combine --target-directory (-t) and --no-target-directory (-T)"
            )
        if destination is None:
            raise click.MissingParameter(
                param_type="argument", param_hint='"SOURCES..."'
            )
        sources = *sources, destination
        target_dir = parse_file_resource(target_directory, root)
        dst = None
    else:
        if destination is None:
            raise click.MissingParameter(
                param_type="argument", param_hint='"DESTINATION"'
            )
        if not sources:
            raise click.MissingParameter(
                param_type="argument", param_hint='"SOURCES..."'
            )
        dst = parse_file_resource(destination, root)
        if no_target_directory or not await _is_dir(root, dst):
            target_dir = None
        else:
            target_dir = dst
            dst = None

    srcs = await _expand(sources, root, glob)
    if no_target_directory and len(srcs) > 1:
        raise click.UsageError(f"Extra operand after {str(srcs[1])!r}")

    errors = False
    for src in srcs:
        if target_dir:
            dst = target_dir / src.name
        assert dst
        try:
            if root.verbosity > 0:
                painter = get_painter(root.color, quote=True)
                src_status = await root.client.storage.stat(src)
            await root.client.storage.mv(src, dst)
        except (OSError, ResourceNotFound, IllegalArgumentError) as error:
            log.error(f"cannot move {src} to {dst}: {error}")
            errors = True
        else:
            if root.verbosity > 0:
                csrc = painter.paint(str(src), src_status.type)
                cdst = painter.paint(str(dst), src_status.type)
                click.echo(f"{csrc} -> {cdst}")

    if errors:
        sys.exit(EX_OSFILE)


@command()
@click.argument("path", required=False)
@option(
    "-a",
    "--all",
    "show_all",
    is_flag=True,
    help="do not ignore entries starting with .",
)
@option(
    "--human-readable",
    "-h",
    is_flag=True,
    help="Print the size in a more human readable way.",
)
@option(
    "--size",
    "-s",
    is_flag=True,
    help="Print the size in bytes of each file.",
)
@option(
    "--sort",
    type=click.Choice(["name", "size", "time"]),
    default="name",
    help="sort by given field, default is name",
)
async def tree(
    root: Root, path: str, size: bool, human_readable: bool, sort: str, show_all: bool
) -> None:
    """List contents of directories in a tree-like format.

    Tree is a recursive directory listing program that produces a depth indented listing
    of files, which is colorized ala dircolors if the LS_COLORS environment variable is
    set and output is to tty.  With no arguments, tree lists the files in the storage:
    directory.  When directory arguments are given, tree lists all the files and/or
    directories found in the given directories each in turn.  Upon completion of listing
    all files/directories found, tree returns the total number of files and/or
    directories listed.

    By default PATH is equal user's home dir (storage:)

    """
    if not path:
        path = "storage:"
    uri = parse_file_resource(path, root)

    errors = False
    try:
        tree = await fetch_tree(root.client, uri, show_all)
        tree = dataclasses.replace(tree, name=str(path))
    except (OSError, ResourceNotFound) as error:
        log.error(f"cannot fetch tree for {uri}: {error}")
        errors = True
    else:
        formatter = TreeFormatter(
            color=root.color, size=size, human_readable=human_readable, sort=sort
        )
        pager_maybe(formatter(tree), root.tty, root.terminal_size)

    if errors:
        sys.exit(EX_OSFILE)


async def _expand(
    paths: Sequence[str], root: Root, glob: bool, allow_file: bool = False
) -> List[URL]:
    uris = []
    for path in paths:
        uri = parse_file_resource(path, root)
        if root.verbosity > 0:
            painter = get_painter(root.color, quote=True)
            curi = painter.paint(str(uri), FileStatusType.FILE)
            click.echo(f"Expand {curi}")
        uri_path = str(_extract_path(uri))
        if glob and globmodule.has_magic(uri_path):
            if uri.scheme == "storage":
                async for file in root.client.storage.glob(uri):
                    uris.append(file)
            elif allow_file and path.startswith("file:"):
                for p in globmodule.iglob(uri_path, recursive=True):
                    uris.append(uri.with_path(p))
            else:
                uris.append(uri)
        else:
            uris.append(uri)
    return uris


async def _is_dir(root: Root, uri: URL) -> bool:
    if uri.scheme == "storage":
        try:
            stat = await root.client.storage.stat(uri)
            return stat.is_dir()
        except ResourceNotFound:
            pass
    elif uri.scheme == "file":
        path = _extract_path(uri)
        return path.is_dir()
    return False


storage.add_command(cp)
storage.add_command(ls)
storage.add_command(glob)
storage.add_command(rm)
storage.add_command(mkdir)
storage.add_command(mv)
storage.add_command(tree)


async def calc_filters(
    client: Client, filters: Optional[Tuple[Tuple[bool, str], ...]]
) -> Tuple[Tuple[bool, str], ...]:
    ret = []
    config = await client.config.get_user_config()
    section = config.get("storage")
    if section is not None:
        for flt in section.get("cp-exclude", ()):
            if flt.startswith("!"):
                ret.append((False, flt[1:]))
            else:
                ret.append((True, flt))
    if filters is not None:
        ret.extend(filters)
    return tuple(ret)


async def calc_ignore_file_names(
    client: Client, exclude_from_files: Optional[str]
) -> List[str]:
    if exclude_from_files is not None:
        return exclude_from_files.split()
    config = await client.config.get_user_config()
    section = config.get("storage")
    ignore_file_names = [NEUROIGNORE_FILENAME]
    if section is not None:
        ignore_file_names = section.get("cp-exclude-from-files", ignore_file_names)
    return ignore_file_names


async def fetch_tree(client: Client, uri: URL, show_all: bool) -> Tree:
    loop = asyncio.get_event_loop()
    folders = []
    files = []
    tasks = []
    size = 0
    async for item in client.storage.ls(uri):
        if not show_all and item.name.startswith("."):
            continue
        if item.is_dir():
            tasks.append(
                loop.create_task(fetch_tree(client, uri / item.name, show_all))
            )
        else:
            files.append(item)
            size += item.size
    for task in tasks:
        subtree = await task
        folders.append(subtree)
        size += subtree.size
    return Tree(uri.name, size, folders, files)



import glob as globmodule  # avoid conflict with subcommand "glob"
import logging
import sys
from itertools import chain
from typing import List, Optional, Sequence, Tuple, Union, cast

import click
from yarl import URL

from neuromation.api import FileStatusType, IllegalArgumentError, ResourceNotFound
from neuromation.api.blob_storage import BlobListing, BucketListing, PrefixListing
from neuromation.api.file_filter import FileFilter
from neuromation.api.url_utils import _extract_path

from .const import EX_OSFILE
from .formatters.blob_storage import (
    BaseBlobFormatter,
    LongBlobFormatter,
    SimpleBlobFormatter,
)
from .formatters.storage import FilesSorter, create_storage_progress, get_painter
from .root import Root
from .storage import calc_filters, calc_ignore_file_names, filter_option
from .utils import (
    command,
    group,
    option,
    pager_maybe,
    parse_blob_or_file_resource,
    parse_blob_resource,
)


log = logging.getLogger(__name__)


@group(name="blob")
def blob_storage() -> None:
    """
    Blob storage operations.
    """


BlobListings = Union[BucketListing, BlobListing, PrefixListing]


@command()
@click.argument("paths", nargs=-1)
@option(
    "--human-readable",
    "-h",
    is_flag=True,
    help="with -l print human readable sizes (e.g., 2K, 540M).",
)
@option("-l", "format_long", is_flag=True, help="use a long listing format.")
@option(
    "--sort",
    type=click.Choice(["name", "size", "time"]),
    default="name",
    help="sort by given field, default is name.",
)
@option(
    "-r",
    "--recursive",
    is_flag=True,
    help="List all keys under the URL path provided, not just 1 level depths.",
)
async def ls(
    root: Root,
    paths: Sequence[str],
    human_readable: bool,
    format_long: bool,
    sort: str,
    recursive: bool,
) -> None:
    """
    List buckets or bucket contents.
    """
    uris = [parse_blob_resource(path, root) for path in paths]

    blob_storage = root.client.blob_storage

    formatter: BaseBlobFormatter
    if format_long:
        # Similar to `ls -l`
        formatter = LongBlobFormatter(human_readable=human_readable, color=root.color)
    else:
        # Similar to `ls -1`, default for non-terminal on UNIX. We show full uris of
        # blobs, thus column formatting does not work too well.
        formatter = SimpleBlobFormatter(root.color)
    sorter = FilesSorter(sort)

    errors = False
    if not uris:
        # List Buckets instead of blobs in bucket
        buckets = await blob_storage.list_buckets()
        pager_maybe(formatter(buckets), root.tty, root.terminal_size)
    else:
        for uri in uris:
            bucket_name, key = blob_storage._extract_bucket_and_key(uri)
            short_uri = blob_storage.make_url(bucket_name, key)
            if root.verbosity > 0:
                painter = get_painter(root.color, quote=True)
                curi = painter.paint(str(short_uri), FileStatusType.DIRECTORY)
                click.echo(f"List of {curi}:")

            try:
                blobs, prefixes = await blob_storage.list_blobs(
                    bucket_name=bucket_name,
                    prefix=key,
                    recursive=recursive,
                )
                items = cast(Sequence[BlobListings], chain(blobs, prefixes))
            except ResourceNotFound as error:
                log.error(f"cannot access {short_uri}: {error}")
                errors = True
            else:
                items = sorted(items, key=sorter.key())
                pager_maybe(formatter(items), root.tty, root.terminal_size)

    if errors:
        sys.exit(EX_OSFILE)


@command()
@click.argument("patterns", nargs=-1, required=False)
async def glob(root: Root, patterns: Sequence[str]) -> None:
    """
    List resources that match PATTERNS.
    """
    blob_storage = root.client.blob_storage
    for pattern in patterns:
        uri = parse_blob_resource(pattern, root)
        bucket_name, pattern = blob_storage._extract_bucket_and_key(uri)
        short_uri = blob_storage.make_url(bucket_name, pattern)

        if globmodule.has_magic(bucket_name):
            raise ValueError(
                "You can not glob on bucket names. Please provide name " "explicitly."
            )

        if root.verbosity > 0:
            painter = get_painter(root.color, quote=True)
            curi = painter.paint(str(short_uri), FileStatusType.FILE)
            click.echo(f"Using pattern {curi}:")

        async for blob in blob_storage.glob_blobs(bucket_name, pattern):
            click.echo(blob.uri)


@command()
@click.argument("sources", nargs=-1, required=False)
@click.argument("destination", required=False)
@option("-r", "--recursive", is_flag=True, help="Recursive copy, off by default")
@option(
    "--glob/--no-glob",
    is_flag=True,
    default=True,
    show_default=True,
    help="Expand glob patterns in SOURCES with explicit scheme.",
)
@option(
    "-t",
    "--target-directory",
    metavar="DIRECTORY",
    default=None,
    help="Copy all SOURCES into DIRECTORY.",
)
@option(
    "-T",
    "--no-target-directory",
    is_flag=True,
    help="Treat DESTINATION as a normal file.",
)
@filter_option(
    "--exclude",
    "filters",
    flag_value=True,
    help=("Exclude files and directories that match the specified pattern."),
)
@filter_option(
    "--include",
    "filters",
    flag_value=False,
    help=("Don't exclude files and directories that match the specified pattern."),
)
@option(
    "--exclude-from-files",
    metavar="FILES",
    default=None,
    help=(
        "A list of file names that contain patterns for exclusion files "
        "and directories. Used only for uploading. "
        "The default can be changed using the storage.cp-exclude-from-files "
        'configuration variable documented in "neuro help user-config"'
    ),
)
@option(
    "-p/-P",
    "--progress/--no-progress",
    is_flag=True,
    default=True,
    help="Show progress, on by default.",
)
async def cp(
    root: Root,
    sources: Sequence[str],
    destination: Optional[str],
    recursive: bool,
    glob: bool,
    target_directory: Optional[str],
    no_target_directory: bool,
    filters: Optional[Tuple[Tuple[bool, str], ...]],
    exclude_from_files: str,
    progress: bool,
) -> None:
    """
    Simple utility to copy files and directories into and from Blob Storage.

    Either SOURCES or DESTINATION should have `blob://` scheme.
    If scheme is omitted, file:// scheme is assumed. It is currently not possible to
    copy files between Blob Storage (`blob://`) destination, nor with `storage://`
    scheme paths.

    Use `/dev/stdin` and `/dev/stdout` file names to upload a file from standard input
    or output to stdout.

    Any number of --exclude and --include options can be passed.  The
    filters that appear later in the command take precedence over filters
    that appear earlier in the command.  If neither --exclude nor
    --include options are specified the default can be changed using the
    storage.cp-exclude configuration variable documented in
    "neuro help user-config".

    File permissions, modification times and other attributes will not be passed to
    Blob Storage metadata during upload.
    """
    target_dir: Optional[URL]
    dst: Optional[URL]
    if target_directory:
        if no_target_directory:
            raise click.UsageError(
                "Cannot combine --target-directory (-t) and --no-target-directory (-T)"
            )
        if destination is None:
            raise click.MissingParameter(
                param_type="argument", param_hint='"SOURCES..."'
            )
        sources = *sources, destination
        target_dir = parse_blob_or_file_resource(target_directory, root)
        dst = None
    else:
        if destination is None:
            raise click.MissingParameter(
                param_type="argument", param_hint='"DESTINATION"'
            )
        if not sources:
            raise click.MissingParameter(
                param_type="argument", param_hint='"SOURCES..."'
            )
        dst = parse_blob_or_file_resource(destination, root)

        # From gsutil:
        #
        # There's an additional wrinkle when working with subdirectories: the resulting
        # names depend on whether the destination subdirectory exists. For example,
        # if gs://my-bucket/subdir exists as a subdirectory, the command:

        # gsutil cp -r dir1/dir2 gs://my-bucket/subdir

        # will create the blob gs://my-bucket/subdir/dir2/a/b/c. In contrast, if
        # gs://my-bucket/subdir does not exist, this same gsutil cp command will create
        # the blob gs://my-bucket/subdir/a/b/c.
        if no_target_directory or not await _is_dir(root, dst):
            target_dir = None
        else:
            target_dir = dst
            dst = None

    ignore_file_names = await calc_ignore_file_names(root.client, exclude_from_files)
    filters = await calc_filters(root.client, filters)
    srcs = await _expand(sources, root, glob, allow_file=True)
    if no_target_directory and len(srcs) > 1:
        raise click.UsageError(f"Extra operand after {str(srcs[1])!r}")

    file_filter = FileFilter()
    for exclude, pattern in filters:
        log.debug("%s %s", "Exclude" if exclude else "Include", pattern)
        file_filter.append(exclude, pattern)

    show_progress = root.tty and progress

    errors = False
    for src in srcs:
        # `src.name` will return empty string if URL has trailing slash, ie.:
        # `neuro blob cp data/ blob:my_bucket` -> dst == blob:my_bucket/file.txt
        # `neuro blob cp data blob:my_bucket` -> dst == blob:my_bucket/data/file.txt
        # `neuro blob cp blob:my_bucket data` -> dst == data/my_bucket/file.txt
        # `neuro blob cp blob:my_bucket/ data` -> dst == data/file.txt
        if target_dir:
            dst = target_dir / src.name
        assert dst

        progress_blob = create_storage_progress(root, show_progress)
        progress_blob.begin(src, dst)

        try:
            if src.scheme == "file" and dst.scheme == "blob":
                if recursive and await _is_dir(root, src):
                    await root.client.blob_storage.upload_dir(
                        src,
                        dst,
                        filter=file_filter.match,
                        ignore_file_names=frozenset(ignore_file_names),
                        progress=progress_blob,
                    )
                else:
                    await root.client.blob_storage.upload_file(
                        src, dst, progress=progress_blob
                    )
            elif src.scheme == "blob" and dst.scheme == "file":
                if recursive and await _is_dir(root, src):
                    await root.client.blob_storage.download_dir(
                        src, dst, filter=file_filter.match, progress=progress_blob
                    )
                else:
                    await root.client.blob_storage.download_file(
                        src, dst, progress=progress_blob
                    )
            else:
                raise RuntimeError(
                    f"Copy operation of the file with scheme '{src.scheme}'"
                    f" to the file with scheme '{dst.scheme}'"
                    f" is not supported"
                )
        except (OSError, ResourceNotFound, IllegalArgumentError) as error:
            log.error(f"cannot copy {src} to {dst}: {error}")
            errors = True

        progress_blob.end()

    if errors:
        sys.exit(EX_OSFILE)


async def _is_dir(root: Root, uri: URL) -> bool:
    if uri.scheme == "blob":
        return await root.client.blob_storage._is_dir(uri)

    elif uri.scheme == "file":
        path = _extract_path(uri)
        return path.is_dir()
    return False


async def _expand(
    paths: Sequence[str], root: Root, glob: bool, allow_file: bool = False
) -> List[URL]:
    uris = []
    for path in paths:
        uri = parse_blob_or_file_resource(path, root)
        if root.verbosity > 0:
            painter = get_painter(root.color, quote=True)
            curi = painter.paint(str(uri), FileStatusType.FILE)
            click.echo(f"Expand {curi}")
        uri_path = str(_extract_path(uri))
        if glob and globmodule.has_magic(uri_path):
            if uri.scheme == "blob":
                bucket_name, key = root.client.blob_storage._extract_bucket_and_key(uri)
                if globmodule.has_magic(bucket_name):
                    raise ValueError(
                        "You can not glob on bucket names. Please provide name "
                        "explicitly."
                    )
                async for blob in root.client.blob_storage.glob_blobs(
                    bucket_name=bucket_name, pattern=key
                ):
                    uris.append(blob.uri)
            elif allow_file and uri.scheme == "file":
                for p in globmodule.iglob(uri_path, recursive=True):
                    uris.append(uri.with_path(p))
            else:
                uris.append(uri)
        else:
            uris.append(uri)
    return uris


blob_storage.add_command(cp)
blob_storage.add_command(ls)
blob_storage.add_command(glob)
